@@ -26,7 +26,6 @@ run_command "jj branch create goodbye"
 run_command "jj log"
 
 comment "Start working on a new change off of master:"
-sleep 1
 run_command "jj co master"
 run_command "jj log"
 
@@ -47,3 +46,5 @@ run_command "jj describe goodbye -m goodbye"
 
 comment "Inspect the result:"
 run_command "jj log"
+
+blank
