@@ -16,7 +16,7 @@
 
 use std::any::Any;
 use std::collections::BTreeMap;
-use std::fmt::{Debug, Error, Formatter};
+use std::fmt::Debug;
 use std::io::Read;
 use std::result::Result;
 use std::vec::Vec;
@@ -24,6 +24,7 @@ use std::vec::Vec;
 use thiserror::Error;
 
 use crate::content_hash::ContentHash;
+use crate::merge::Merge;
 use crate::repo_path::{RepoPath, RepoPathComponent};
 
 pub trait ObjectId {
@@ -42,14 +43,14 @@ macro_rules! id_type {
             #[derive(PartialEq, Eq, PartialOrd, Ord, Clone, Hash)]
             $vis struct $name(Vec<u8>);
         }
-        impl_id_type!($name);
+        $crate::backend::impl_id_type!($name);
     };
 }
 
 macro_rules! impl_id_type {
     ($name:ident) => {
-        impl Debug for $name {
-            fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), Error> {
+        impl std::fmt::Debug for $name {
+            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
                 f.debug_tuple(stringify!($name)).field(&self.hex()).finish()
             }
         }
@@ -90,6 +91,8 @@ macro_rules! impl_id_type {
     };
 }
 
+pub(crate) use {id_type, impl_id_type};
+
 id_type!(pub CommitId);
 id_type!(pub ChangeId);
 id_type!(pub TreeId);
@@ -140,12 +143,65 @@ content_hash! {
     }
 }
 
+/// Identifies a single legacy tree, which may have path-level conflicts, or a
+/// merge of multiple trees, where the individual trees do not have conflicts.
+// TODO(#1624): Delete this type at some point in the future, when we decide to drop
+// support for conflicts in older repos, or maybe after we have provided an upgrade
+// mechanism.
+#[derive(Debug, Clone)]
+pub enum MergedTreeId {
+    /// The tree id of a legacy tree
+    Legacy(TreeId),
+    /// The tree id(s) of a merge tree
+    Merge(Merge<TreeId>),
+}
+
+impl PartialEq for MergedTreeId {
+    /// Overridden to make conflict-free trees be considered equal even if their
+    /// `MergedTreeId` variant is different.
+    fn eq(&self, other: &Self) -> bool {
+        self.to_merge() == other.to_merge()
+    }
+}
+
+impl Eq for MergedTreeId {}
+
+impl ContentHash for MergedTreeId {
+    fn hash(&self, state: &mut impl digest::Update) {
+        match self {
+            MergedTreeId::Legacy(tree_id) => {
+                state.update(b"0");
+                ContentHash::hash(tree_id, state);
+            }
+            MergedTreeId::Merge(tree_ids) => {
+                state.update(b"1");
+                ContentHash::hash(tree_ids, state);
+            }
+        }
+    }
+}
+
+impl MergedTreeId {
+    /// Create a resolved `MergedTreeId` from a single regular tree.  
+    pub fn resolved(tree_id: TreeId) -> Self {
+        MergedTreeId::Merge(Merge::resolved(tree_id))
+    }
+
+    /// Return this id as `Merge<TreeId>`
+    pub fn to_merge(&self) -> Merge<TreeId> {
+        match self {
+            MergedTreeId::Legacy(tree_id) => Merge::resolved(tree_id.clone()),
+            MergedTreeId::Merge(tree_ids) => tree_ids.clone(),
+        }
+    }
+}
+
 content_hash! {
     #[derive(Debug, PartialEq, Eq, Clone)]
     pub struct Commit {
         pub parents: Vec<CommitId>,
         pub predecessors: Vec<CommitId>,
-        pub root_tree: TreeId,
+        pub root_tree: MergedTreeId,
         pub change_id: ChangeId,
         pub description: String,
         pub author: Signature,
@@ -205,7 +261,9 @@ pub enum BackendError {
     InvalidUtf8 {
         object_type: String,
         hash: String,
-        source: std::string::FromUtf8Error,
+        // Box to reduce size for other error types that include this one:
+        // https://rust-lang.github.io/rust-clippy/master/index.html#result_large_err
+        source: Box<std::string::FromUtf8Error>,
     },
     #[error("Object {hash} of type {object_type} not found: {source}")]
     ObjectNotFound {
@@ -239,6 +297,18 @@ pub enum TreeValue {
     Conflict(ConflictId),
 }
 
+impl TreeValue {
+    pub fn hex(&self) -> String {
+        match self {
+            TreeValue::File { id, .. } => id.hex(),
+            TreeValue::Symlink(id) => id.hex(),
+            TreeValue::Tree(id) => id.hex(),
+            TreeValue::GitSubmodule(id) => id.hex(),
+            TreeValue::Conflict(id) => id.hex(),
+        }
+    }
+}
+
 impl ContentHash for TreeValue {
     fn hash(&self, state: &mut impl digest::Update) {
         use TreeValue::*;
@@ -387,7 +457,7 @@ pub fn make_root_commit(root_change_id: ChangeId, empty_tree_id: TreeId) -> Comm
     Commit {
         parents: vec![],
         predecessors: vec![],
-        root_tree: empty_tree_id,
+        root_tree: MergedTreeId::Legacy(empty_tree_id),
         change_id: root_change_id,
         description: String::new(),
         author: signature.clone(),
