@@ -20,10 +20,9 @@ use std::hash::{Hash, Hasher};
 use std::sync::Arc;
 
 use crate::backend;
-use crate::backend::{ChangeId, CommitId, Signature, TreeId};
-use crate::repo_path::RepoPath;
+use crate::backend::{BackendError, ChangeId, CommitId, MergedTreeId, Signature};
+use crate::merged_tree::MergedTree;
 use crate::store::Store;
-use crate::tree::Tree;
 
 #[derive(Clone)]
 pub struct Commit {
@@ -54,7 +53,7 @@ impl Ord for Commit {
 
 impl PartialOrd for Commit {
     fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
-        Some(self.id.cmp(&other.id))
+        Some(self.cmp(other))
     }
 }
 
@@ -101,13 +100,11 @@ impl Commit {
             .collect()
     }
 
-    pub fn tree(&self) -> Tree {
-        self.store
-            .get_tree(&RepoPath::root(), &self.data.root_tree)
-            .unwrap()
+    pub fn tree(&self) -> Result<MergedTree, BackendError> {
+        self.store.get_root_tree(&self.data.root_tree)
     }
 
-    pub fn tree_id(&self) -> &TreeId {
+    pub fn tree_id(&self) -> &MergedTreeId {
         &self.data.root_tree
     }
 
@@ -142,3 +139,23 @@ impl Commit {
         false
     }
 }
+
+/// Wrapper to sort `Commit` by committer timestamp.
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
+pub(crate) struct CommitByCommitterTimestamp(pub Commit);
+
+impl Ord for CommitByCommitterTimestamp {
+    fn cmp(&self, other: &Self) -> Ordering {
+        let self_timestamp = &self.0.committer().timestamp.timestamp;
+        let other_timestamp = &other.0.committer().timestamp.timestamp;
+        self_timestamp
+            .cmp(other_timestamp)
+            .then_with(|| self.0.cmp(&other.0)) // to comply with Eq
+    }
+}
+
+impl PartialOrd for CommitByCommitterTimestamp {
+    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
+        Some(self.cmp(other))
+    }
+}
