@@ -15,7 +15,7 @@
 #![allow(missing_docs)]
 
 use std::cmp::{Ordering, Reverse};
-use std::collections::{BinaryHeap, HashSet};
+use std::collections::{BTreeSet, BinaryHeap, HashSet};
 use std::fmt;
 use std::iter::Peekable;
 use std::ops::Range;
@@ -29,13 +29,14 @@ use crate::default_index_store::{
 };
 use crate::default_revset_graph_iterator::RevsetGraphIterator;
 use crate::id_prefix::{IdIndex, IdIndexSource, IdIndexSourceEntry};
-use crate::index::{HexPrefix, Index, PrefixResolution};
+use crate::index::{HexPrefix, PrefixResolution};
 use crate::matchers::{EverythingMatcher, Matcher, PrefixMatcher, Visit};
 use crate::repo_path::RepoPath;
 use crate::revset::{
     ChangeIdIndex, ResolvedExpression, ResolvedPredicateExpression, Revset, RevsetEvaluationError,
-    RevsetFilterPredicate, RevsetGraphEdge, GENERATION_RANGE_FULL,
+    RevsetFilterPredicate, GENERATION_RANGE_FULL,
 };
+use crate::revset_graph::RevsetGraphEdge;
 use crate::rewrite;
 use crate::store::Store;
 
@@ -78,7 +79,7 @@ impl<'index> RevsetImpl<'index> {
     }
 
     pub fn iter_graph_impl(&self) -> RevsetGraphIterator<'_, 'index> {
-        RevsetGraphIterator::new(self.inner.iter())
+        RevsetGraphIterator::new(self.index, self.inner.iter())
     }
 }
 
@@ -104,7 +105,7 @@ impl<'index> Revset<'index> for RevsetImpl<'index> {
     }
 
     fn iter_graph(&self) -> Box<dyn Iterator<Item = (CommitId, Vec<RevsetGraphEdge>)> + '_> {
-        Box::new(RevsetGraphIterator::new(self.inner.iter()))
+        Box::new(self.iter_graph_impl())
     }
 
     fn change_id_index(&self) -> Box<dyn ChangeIdIndex + 'index> {
@@ -593,13 +594,15 @@ impl<'index> EvaluationContext<'index> {
             }
             ResolvedExpression::Heads(candidates) => {
                 let candidate_set = self.evaluate(candidates)?;
-                let candidate_ids = candidate_set
-                    .iter()
-                    .map(|entry| entry.commit_id())
-                    .collect_vec();
-                Ok(Box::new(self.revset_for_commit_ids(
-                    &self.index.heads(&mut candidate_ids.iter()),
-                )))
+                let head_positions: BTreeSet<_> = self
+                    .index
+                    .heads_pos(candidate_set.iter().map(|entry| entry.position()).collect());
+                let index_entries = head_positions
+                    .into_iter()
+                    .rev()
+                    .map(|pos| self.index.entry_by_pos(pos))
+                    .collect();
+                Ok(Box::new(EagerRevset { index_entries }))
             }
             ResolvedExpression::Roots(candidates) => {
                 let candidate_set = EagerRevset {
@@ -711,7 +714,7 @@ impl<'index> EvaluationContext<'index> {
         }
     }
 
-    /// Calculates `root_set:head_set`.
+    /// Calculates `root_set::head_set`.
     fn collect_dag_range<'a, 'b, S, T>(
         &self,
         root_set: &S,
@@ -828,33 +831,29 @@ fn build_predicate_fn<'index>(
             let parent_count_range = parent_count_range.clone();
             pure_predicate_fn(move |entry| parent_count_range.contains(&entry.num_parents()))
         }
-        RevsetFilterPredicate::Description(needle) => {
-            let needle = needle.clone();
+        RevsetFilterPredicate::Description(pattern) => {
+            let pattern = pattern.clone();
             pure_predicate_fn(move |entry| {
-                store
-                    .get_commit(&entry.commit_id())
-                    .unwrap()
-                    .description()
-                    .contains(needle.as_str())
+                let commit = store.get_commit(&entry.commit_id()).unwrap();
+                pattern.matches(commit.description())
             })
         }
-        RevsetFilterPredicate::Author(needle) => {
-            let needle = needle.clone();
+        RevsetFilterPredicate::Author(pattern) => {
+            let pattern = pattern.clone();
             // TODO: Make these functions that take a needle to search for accept some
             // syntax for specifying whether it's a regex and whether it's
             // case-sensitive.
             pure_predicate_fn(move |entry| {
                 let commit = store.get_commit(&entry.commit_id()).unwrap();
-                commit.author().name.contains(needle.as_str())
-                    || commit.author().email.contains(needle.as_str())
+                pattern.matches(&commit.author().name) || pattern.matches(&commit.author().email)
             })
         }
-        RevsetFilterPredicate::Committer(needle) => {
-            let needle = needle.clone();
+        RevsetFilterPredicate::Committer(pattern) => {
+            let pattern = pattern.clone();
             pure_predicate_fn(move |entry| {
                 let commit = store.get_commit(&entry.commit_id()).unwrap();
-                commit.committer().name.contains(needle.as_str())
-                    || commit.committer().email.contains(needle.as_str())
+                pattern.matches(&commit.committer().name)
+                    || pattern.matches(&commit.committer().email)
             })
         }
         RevsetFilterPredicate::File(paths) => {
@@ -870,7 +869,7 @@ fn build_predicate_fn<'index>(
         }
         RevsetFilterPredicate::HasConflict => pure_predicate_fn(move |entry| {
             let commit = store.get_commit(&entry.commit_id()).unwrap();
-            commit.tree().has_conflict()
+            commit.tree().unwrap().has_conflict()
         }),
     }
 }
@@ -893,7 +892,7 @@ fn has_diff_from_parent(
         }
     }
     let from_tree = rewrite::merge_commit_trees_without_repo(store, &index, &parents).unwrap();
-    let to_tree = commit.tree();
+    let to_tree = commit.tree().unwrap();
     from_tree.diff(&to_tree, matcher).next().is_some()
 }
 
