@@ -17,6 +17,7 @@
 use std::any::Any;
 use std::fmt::{Debug, Error, Formatter};
 use std::io::{Cursor, Read};
+use std::ops::Deref;
 use std::path::Path;
 use std::sync::{Arc, Mutex, MutexGuard};
 use std::{fs, slice};
@@ -28,11 +29,12 @@ use thiserror::Error;
 
 use crate::backend::{
     make_root_commit, Backend, BackendError, BackendInitError, BackendLoadError, BackendResult,
-    ChangeId, Commit, CommitId, Conflict, ConflictId, ConflictTerm, FileId, MillisSinceEpoch,
-    ObjectId, Signature, SymlinkId, Timestamp, Tree, TreeId, TreeValue,
+    ChangeId, Commit, CommitId, Conflict, ConflictId, ConflictTerm, FileId, MergedTreeId,
+    MillisSinceEpoch, ObjectId, Signature, SymlinkId, Timestamp, Tree, TreeId, TreeValue,
 };
 use crate::file_util::{IoResultExt as _, PathError};
 use crate::lock::FileLock;
+use crate::merge::{Merge, MergeBuilder};
 use crate::repo_path::{RepoPath, RepoPathComponent};
 use crate::stacked_table::{
     MutableTable, ReadonlyTable, TableSegment, TableStore, TableStoreError,
@@ -219,6 +221,9 @@ fn commit_from_git_without_root_parent(commit: &git2::Commit) -> Commit {
         .map(|oid| CommitId::from_bytes(oid.as_bytes()))
         .collect_vec();
     let tree_id = TreeId::from_bytes(commit.tree_id().as_bytes());
+    // If this commit is a conflict, we'll update the root tree later, when we read
+    // the extra metadata.
+    let root_tree = MergedTreeId::Legacy(tree_id);
     let description = commit.message().unwrap_or("<no message>").to_owned();
     let author = signature_from_git(commit.author());
     let committer = signature_from_git(commit.committer());
@@ -226,7 +231,8 @@ fn commit_from_git_without_root_parent(commit: &git2::Commit) -> Commit {
     Commit {
         parents,
         predecessors: vec![],
-        root_tree: tree_id,
+        // If this commit has associated extra metadata, we may reset this later.
+        root_tree,
         change_id,
         description,
         author,
@@ -234,9 +240,21 @@ fn commit_from_git_without_root_parent(commit: &git2::Commit) -> Commit {
     }
 }
 
+const EMPTY_STRING_PLACEHOLDER: &str = "JJ_EMPTY_STRING";
+
 fn signature_from_git(signature: git2::Signature) -> Signature {
-    let name = signature.name().unwrap_or("<no name>").to_owned();
-    let email = signature.email().unwrap_or("<no email>").to_owned();
+    let name = signature.name().unwrap_or_default();
+    let name = if name != EMPTY_STRING_PLACEHOLDER {
+        name.to_owned()
+    } else {
+        "".to_string()
+    };
+    let email = signature.email().unwrap_or_default();
+    let email = if email != EMPTY_STRING_PLACEHOLDER {
+        email.to_owned()
+    } else {
+        "".to_string()
+    };
     let timestamp = MillisSinceEpoch(signature.when().seconds() * 1000);
     let tz_offset = signature.when().offset_minutes();
     Signature {
@@ -250,8 +268,17 @@ fn signature_from_git(signature: git2::Signature) -> Signature {
 }
 
 fn signature_to_git(signature: &Signature) -> git2::Signature<'static> {
-    let name = &signature.name;
-    let email = &signature.email;
+    // git does not support empty names or emails
+    let name = if !signature.name.is_empty() {
+        &signature.name
+    } else {
+        EMPTY_STRING_PLACEHOLDER
+    };
+    let email = if !signature.email.is_empty() {
+        &signature.email
+    } else {
+        EMPTY_STRING_PLACEHOLDER
+    };
     let time = git2::Time::new(
         signature.timestamp.timestamp.0.div_euclid(1000),
         signature.timestamp.tz_offset,
@@ -264,6 +291,12 @@ fn serialize_extras(commit: &Commit) -> Vec<u8> {
         change_id: commit.change_id.to_bytes(),
         ..Default::default()
     };
+    if let MergedTreeId::Merge(tree_ids) = &commit.root_tree {
+        proto.uses_tree_conflict_format = true;
+        if !tree_ids.is_resolved() {
+            proto.root_tree = tree_ids.iter().map(|r| r.to_bytes()).collect();
+        }
+    }
     for predecessor in &commit.predecessors {
         proto.predecessors.push(predecessor.to_bytes());
     }
@@ -273,6 +306,24 @@ fn serialize_extras(commit: &Commit) -> Vec<u8> {
 fn deserialize_extras(commit: &mut Commit, bytes: &[u8]) {
     let proto = crate::protos::git_store::Commit::decode(bytes).unwrap();
     commit.change_id = ChangeId::new(proto.change_id);
+    if proto.uses_tree_conflict_format {
+        if !proto.root_tree.is_empty() {
+            let merge_builder: MergeBuilder<_> = proto
+                .root_tree
+                .iter()
+                .map(|id_bytes| TreeId::from_bytes(id_bytes))
+                .collect();
+            commit.root_tree = MergedTreeId::Merge(merge_builder.build());
+        } else {
+            // uses_tree_conflict_format was set but there was no root_tree override in the
+            // proto, which means we should just promote the tree id from the
+            // git commit to be a known-conflict-free tree
+            let MergedTreeId::Legacy(legacy_tree_id) = &commit.root_tree else {
+                panic!("root tree should have been initialized to a legacy id");
+            };
+            commit.root_tree = MergedTreeId::resolved(legacy_tree_id.clone());
+        }
+    }
     for predecessor in &proto.predecessors {
         commit.predecessors.push(CommitId::from_bytes(predecessor));
     }
@@ -403,7 +454,7 @@ impl Backend for GitBackend {
             BackendError::InvalidUtf8 {
                 object_type: id.object_type(),
                 hash: id.hex(),
-                source: err,
+                source: Box::new(err),
             }
         })?;
         Ok(target)
@@ -597,6 +648,9 @@ impl Backend for GitBackend {
                 // the commit is a branch head, so bulk-import metadata as much as possible.
                 tracing::debug!("import extra metadata entries");
                 let mut mut_table = table.start_mutation();
+                // TODO(#1624): Should we read the root tree here and check if it has a
+                // `.jjconflict-...` entries? That could happen if the user used `git` to e.g.
+                // change the description of a commit with tree-level conflicts.
                 mut_table.add_entry(id.to_bytes(), serialize_extras(&commit));
                 if commit.parents != slice::from_ref(&self.root_commit_id) {
                     import_extra_metadata_entries_from_heads(
@@ -615,10 +669,16 @@ impl Backend for GitBackend {
 
     fn write_commit(&self, mut contents: Commit) -> BackendResult<(CommitId, Commit)> {
         let locked_repo = self.repo.lock().unwrap();
-        let git_tree_id = validate_git_object_id(&contents.root_tree)?;
+        let git_tree_id = match &contents.root_tree {
+            MergedTreeId::Legacy(tree_id) => validate_git_object_id(tree_id)?,
+            MergedTreeId::Merge(tree_ids) => match tree_ids.as_resolved() {
+                Some(tree_id) => validate_git_object_id(tree_id)?,
+                None => write_tree_conflict(locked_repo.deref(), tree_ids)?,
+            },
+        };
         let git_tree = locked_repo
             .find_tree(git_tree_id)
-            .map_err(|err| map_not_found_err(err, &contents.root_tree))?;
+            .map_err(|err| map_not_found_err(err, &TreeId::from_bytes(git_tree_id.as_bytes())))?;
         let author = signature_to_git(&contents.author);
         let mut committer = signature_to_git(&contents.committer);
         let message = &contents.description;
@@ -705,6 +765,29 @@ impl Backend for GitBackend {
     }
 }
 
+/// Write a tree conflict as a special tree with `.jjconflict-base-N` and
+/// `.jjconflict-base-N` subtrees. This ensure that the parts are not GC'd.
+fn write_tree_conflict(
+    repo: &git2::Repository,
+    conflict: &Merge<TreeId>,
+) -> Result<Oid, BackendError> {
+    let mut builder = repo.treebuilder(None).unwrap();
+    let mut add_tree_entry = |name, tree_id: &TreeId| {
+        let tree_oid = Oid::from_bytes(tree_id.as_bytes()).unwrap();
+        builder.insert(name, tree_oid, 0o040000).unwrap();
+    };
+    for (i, tree_id) in conflict.removes().iter().enumerate() {
+        add_tree_entry(format!(".jjconflict-base-{i}"), tree_id);
+    }
+    for (i, tree_id) in conflict.adds().iter().enumerate() {
+        add_tree_entry(format!(".jjconflict-side-{i}"), tree_id);
+    }
+    builder.write().map_err(|err| BackendError::WriteObject {
+        object_type: "tree",
+        source: Box::new(err),
+    })
+}
+
 fn conflict_term_list_to_json(parts: &[ConflictTerm]) -> serde_json::Value {
     serde_json::Value::Array(parts.iter().map(conflict_term_to_json).collect())
 }
@@ -836,7 +919,10 @@ mod tests {
         assert_eq!(&commit.change_id, &change_id);
         assert_eq!(commit.parents, vec![CommitId::from_bytes(&[0; 20])]);
         assert_eq!(commit.predecessors, vec![]);
-        assert_eq!(commit.root_tree.as_bytes(), root_tree_id.as_bytes());
+        assert_eq!(
+            commit.root_tree,
+            MergedTreeId::Legacy(TreeId::from_bytes(root_tree_id.as_bytes()))
+        );
         assert_eq!(commit.description, "git commit message");
         assert_eq!(commit.author.name, "git author");
         assert_eq!(commit.author.email, "git.author@example.com");
@@ -893,6 +979,54 @@ mod tests {
         );
     }
 
+    #[test]
+    fn read_empty_string_placeholder() {
+        let git_signature1 = git2::Signature::new(
+            EMPTY_STRING_PLACEHOLDER,
+            "git.author@example.com",
+            &git2::Time::new(1000, 60),
+        )
+        .unwrap();
+        let signature1 = signature_from_git(git_signature1);
+        assert!(signature1.name.is_empty());
+        assert_eq!(signature1.email, "git.author@example.com");
+        let git_signature2 = git2::Signature::new(
+            "git committer",
+            EMPTY_STRING_PLACEHOLDER,
+            &git2::Time::new(2000, -480),
+        )
+        .unwrap();
+        let signature2 = signature_from_git(git_signature2);
+        assert_eq!(signature2.name, "git committer");
+        assert!(signature2.email.is_empty());
+    }
+
+    #[test]
+    fn write_empty_string_placeholder() {
+        let signature1 = Signature {
+            name: "".to_string(),
+            email: "someone@example.com".to_string(),
+            timestamp: Timestamp {
+                timestamp: MillisSinceEpoch(0),
+                tz_offset: 0,
+            },
+        };
+        let git_signature1 = signature_to_git(&signature1);
+        assert_eq!(git_signature1.name().unwrap(), EMPTY_STRING_PLACEHOLDER);
+        assert_eq!(git_signature1.email().unwrap(), "someone@example.com");
+        let signature2 = Signature {
+            name: "Someone".to_string(),
+            email: "".to_string(),
+            timestamp: Timestamp {
+                timestamp: MillisSinceEpoch(0),
+                tz_offset: 0,
+            },
+        };
+        let git_signature2 = signature_to_git(&signature2);
+        assert_eq!(git_signature2.name().unwrap(), "Someone");
+        assert_eq!(git_signature2.email().unwrap(), EMPTY_STRING_PLACEHOLDER);
+    }
+
     /// Test that parents get written correctly
     #[test]
     fn git_commit_parents() {
@@ -905,7 +1039,7 @@ mod tests {
         let mut commit = Commit {
             parents: vec![],
             predecessors: vec![],
-            root_tree: backend.empty_tree_id().clone(),
+            root_tree: MergedTreeId::Legacy(backend.empty_tree_id().clone()),
             change_id: ChangeId::from_hex("abc123"),
             description: "".to_string(),
             author: create_signature(),
@@ -957,6 +1091,80 @@ mod tests {
         );
     }
 
+    #[test]
+    fn write_tree_conflicts() {
+        let temp_dir = testutils::new_temp_dir();
+        let store_path = temp_dir.path();
+        let git_repo_path = temp_dir.path().join("git");
+        let git_repo = git2::Repository::init(&git_repo_path).unwrap();
+
+        let backend = GitBackend::init_external(store_path, &git_repo_path).unwrap();
+        let create_tree = |i| {
+            let blob_id = git_repo.blob(b"content {i}").unwrap();
+            let mut tree_builder = git_repo.treebuilder(None).unwrap();
+            tree_builder
+                .insert(format!("file{i}"), blob_id, 0o100644)
+                .unwrap();
+            TreeId::from_bytes(tree_builder.write().unwrap().as_bytes())
+        };
+
+        let root_tree = Merge::new(
+            vec![create_tree(0), create_tree(1)],
+            vec![create_tree(2), create_tree(3), create_tree(4)],
+        );
+        let mut commit = Commit {
+            parents: vec![backend.root_commit_id().clone()],
+            predecessors: vec![],
+            root_tree: MergedTreeId::Merge(root_tree.clone()),
+            change_id: ChangeId::from_hex("abc123"),
+            description: "".to_string(),
+            author: create_signature(),
+            committer: create_signature(),
+        };
+
+        // When writing a tree-level conflict, the root tree on the git side has the
+        // individual trees as subtrees.
+        let read_commit_id = backend.write_commit(commit.clone()).unwrap().0;
+        let read_commit = backend.read_commit(&read_commit_id).unwrap();
+        assert_eq!(read_commit, commit);
+        let git_commit = git_repo
+            .find_commit(Oid::from_bytes(read_commit_id.as_bytes()).unwrap())
+            .unwrap();
+        let git_tree = git_repo.find_tree(git_commit.tree_id()).unwrap();
+        assert!(git_tree.iter().all(|entry| entry.filemode() == 0o040000));
+        let mut iter = git_tree.iter();
+        let entry = iter.next().unwrap();
+        assert_eq!(entry.name(), Some(".jjconflict-base-0"));
+        assert_eq!(entry.id().as_bytes(), root_tree.removes()[0].as_bytes());
+        let entry = iter.next().unwrap();
+        assert_eq!(entry.name(), Some(".jjconflict-base-1"));
+        assert_eq!(entry.id().as_bytes(), root_tree.removes()[1].as_bytes());
+        let entry = iter.next().unwrap();
+        assert_eq!(entry.name(), Some(".jjconflict-side-0"));
+        assert_eq!(entry.id().as_bytes(), root_tree.adds()[0].as_bytes());
+        let entry = iter.next().unwrap();
+        assert_eq!(entry.name(), Some(".jjconflict-side-1"));
+        assert_eq!(entry.id().as_bytes(), root_tree.adds()[1].as_bytes());
+        let entry = iter.next().unwrap();
+        assert_eq!(entry.name(), Some(".jjconflict-side-2"));
+        assert_eq!(entry.id().as_bytes(), root_tree.adds()[2].as_bytes());
+        assert!(iter.next().is_none());
+
+        // When writing a single tree using the new format, it's represented by a
+        // regular git tree.
+        commit.root_tree = MergedTreeId::resolved(create_tree(5));
+        let read_commit_id = backend.write_commit(commit.clone()).unwrap().0;
+        let read_commit = backend.read_commit(&read_commit_id).unwrap();
+        assert_eq!(read_commit, commit);
+        let git_commit = git_repo
+            .find_commit(Oid::from_bytes(read_commit_id.as_bytes()).unwrap())
+            .unwrap();
+        assert_eq!(
+            MergedTreeId::resolved(TreeId::from_bytes(git_commit.tree_id().as_bytes())),
+            commit.root_tree
+        );
+    }
+
     #[test]
     fn commit_has_ref() {
         let temp_dir = testutils::new_temp_dir();
@@ -972,7 +1180,7 @@ mod tests {
         let commit = Commit {
             parents: vec![store.root_commit_id().clone()],
             predecessors: vec![],
-            root_tree: store.empty_tree_id().clone(),
+            root_tree: MergedTreeId::Legacy(store.empty_tree_id().clone()),
             change_id: ChangeId::new(vec![]),
             description: "initial".to_string(),
             author: signature.clone(),
@@ -995,7 +1203,7 @@ mod tests {
         let mut commit1 = Commit {
             parents: vec![store.root_commit_id().clone()],
             predecessors: vec![],
-            root_tree: store.empty_tree_id().clone(),
+            root_tree: MergedTreeId::Legacy(store.empty_tree_id().clone()),
             change_id: ChangeId::new(vec![]),
             description: "initial".to_string(),
             author: create_signature(),
