@@ -32,6 +32,7 @@ pub mod default_submodule_store;
 pub mod diff;
 pub mod file_util;
 pub mod files;
+pub mod fmt_util;
 pub mod fsmonitor;
 pub mod git;
 pub mod git_backend;
@@ -43,7 +44,7 @@ pub mod local_backend;
 pub mod lock;
 pub mod matchers;
 pub mod merge;
-pub mod nightly_shims;
+pub mod merged_tree;
 pub mod op_heads_store;
 pub mod op_store;
 pub mod operation;
@@ -53,6 +54,7 @@ pub mod refs;
 pub mod repo;
 pub mod repo_path;
 pub mod revset;
+pub mod revset_graph;
 pub mod rewrite;
 pub mod settings;
 pub mod simple_op_heads_store;
