@@ -16,23 +16,24 @@
 
 use std::collections::{HashMap, HashSet};
 use std::fmt::{Debug, Formatter};
-use std::fs;
 use std::io::ErrorKind;
 use std::ops::Deref;
 use std::path::{Path, PathBuf};
 use std::pin::Pin;
 use std::sync::Arc;
+use std::{fs, slice};
 
 use itertools::Itertools;
 use once_cell::sync::OnceCell;
 use thiserror::Error;
+use tracing::instrument;
 
 use self::dirty_cell::DirtyCell;
 use crate::backend::{
     Backend, BackendError, BackendInitError, BackendLoadError, BackendResult, ChangeId, CommitId,
-    ObjectId, TreeId,
+    MergedTreeId, ObjectId,
 };
-use crate::commit::Commit;
+use crate::commit::{Commit, CommitByCommitterTimestamp};
 use crate::commit_builder::CommitBuilder;
 use crate::default_index_store::DefaultIndexStore;
 use crate::default_submodule_store::DefaultSubmoduleStore;
@@ -41,7 +42,7 @@ use crate::git_backend::GitBackend;
 use crate::index::{HexPrefix, Index, IndexStore, MutableIndex, PrefixResolution, ReadonlyIndex};
 use crate::local_backend::LocalBackend;
 use crate::op_heads_store::{self, OpHeadResolutionError, OpHeadsStore};
-use crate::op_store::{BranchTarget, OpStore, OperationId, RefTarget, WorkspaceId};
+use crate::op_store::{BranchTarget, OpStore, OpStoreError, OperationId, RefTarget, WorkspaceId};
 use crate::operation::Operation;
 use crate::refs::merge_ref_targets;
 use crate::revset::{self, ChangeIdIndex, Revset, RevsetExpression};
@@ -151,7 +152,7 @@ impl ReadonlyRepo {
         let backend = backend_factory(&store_path)?;
         let backend_path = store_path.join("type");
         fs::write(&backend_path, backend.name()).context(&backend_path)?;
-        let store = Store::new(backend);
+        let store = Store::new(backend, user_settings.use_tree_conflict_format());
         let repo_settings = user_settings.with_repo(&repo_path).unwrap();
 
         let op_store_path = repo_path.join("op_store");
@@ -298,11 +299,12 @@ impl ReadonlyRepo {
     pub fn reload_at_head(
         &self,
         user_settings: &UserSettings,
-    ) -> Result<Arc<ReadonlyRepo>, OpHeadResolutionError<TreeMergeError>> {
+    ) -> Result<Arc<ReadonlyRepo>, OpHeadResolutionError<RepoLoaderError>> {
         self.loader().load_at_head(user_settings)
     }
 
-    pub fn reload_at(&self, operation: &Operation) -> Arc<ReadonlyRepo> {
+    #[instrument]
+    pub fn reload_at(&self, operation: &Operation) -> Result<Arc<ReadonlyRepo>, RepoLoaderError> {
         self.loader().load_at(operation)
     }
 }
@@ -558,6 +560,14 @@ fn read_store_type_compat(
         .map_err(|source| StoreLoadError::ReadError { store, source })
 }
 
+#[derive(Debug, Error)]
+pub enum RepoLoaderError {
+    #[error(transparent)]
+    TreeMerge(#[from] TreeMergeError),
+    #[error(transparent)]
+    OpStore(#[from] OpStoreError),
+}
+
 #[derive(Clone)]
 pub struct RepoLoader {
     repo_path: PathBuf,
@@ -575,7 +585,10 @@ impl RepoLoader {
         repo_path: &Path,
         store_factories: &StoreFactories,
     ) -> Result<Self, StoreLoadError> {
-        let store = Store::new(store_factories.load_backend(&repo_path.join("store"))?);
+        let store = Store::new(
+            store_factories.load_backend(&repo_path.join("store"))?,
+            user_settings.use_tree_conflict_format(),
+        );
         let repo_settings = user_settings.with_repo(repo_path).unwrap();
         let op_store = Arc::from(store_factories.load_op_store(&repo_path.join("op_store"))?);
         let op_heads_store =
@@ -617,19 +630,20 @@ impl RepoLoader {
     pub fn load_at_head(
         &self,
         user_settings: &UserSettings,
-    ) -> Result<Arc<ReadonlyRepo>, OpHeadResolutionError<TreeMergeError>> {
+    ) -> Result<Arc<ReadonlyRepo>, OpHeadResolutionError<RepoLoaderError>> {
         let op = op_heads_store::resolve_op_heads(
             self.op_heads_store.as_ref(),
             &self.op_store,
             |op_heads| self._resolve_op_heads(op_heads, user_settings),
         )?;
-        let view = View::new(op.view().take_store_view());
+        let view = View::new(op.view()?.take_store_view());
         Ok(self._finish_load(op, view))
     }
 
-    pub fn load_at(&self, op: &Operation) -> Arc<ReadonlyRepo> {
-        let view = View::new(op.view().take_store_view());
-        self._finish_load(op.clone(), view)
+    #[instrument(skip(self))]
+    pub fn load_at(&self, op: &Operation) -> Result<Arc<ReadonlyRepo>, RepoLoaderError> {
+        let view = View::new(op.view()?.take_store_view());
+        Ok(self._finish_load(op.clone(), view))
     }
 
     pub fn create_from(
@@ -658,11 +672,11 @@ impl RepoLoader {
         &self,
         op_heads: Vec<Operation>,
         user_settings: &UserSettings,
-    ) -> Result<Operation, TreeMergeError> {
-        let base_repo = self.load_at(&op_heads[0]);
+    ) -> Result<Operation, RepoLoaderError> {
+        let base_repo = self.load_at(&op_heads[0])?;
         let mut tx = base_repo.start_transaction(user_settings, "resolve concurrent operations");
         for other_op_head in op_heads.into_iter().skip(1) {
-            tx.merge_operation(other_op_head);
+            tx.merge_operation(other_op_head)?;
             tx.mut_repo().rebase_descendants(user_settings)?;
         }
         let merged_repo = tx.write().leave_unpublished();
@@ -739,7 +753,7 @@ impl MutableRepo {
         &mut self,
         settings: &UserSettings,
         parents: Vec<CommitId>,
-        tree_id: TreeId,
+        tree_id: MergedTreeId,
     ) -> CommitBuilder {
         CommitBuilder::for_new_commit(self, settings, parents, tree_id)
     }
@@ -860,8 +874,7 @@ impl MutableRepo {
         fn local_branch_target_ids(view: &View) -> impl Iterator<Item = &CommitId> {
             view.branches()
                 .values()
-                .filter_map(|branch_target| branch_target.local_target.as_ref())
-                .flat_map(|target| target.adds())
+                .flat_map(|branch_target| branch_target.local_target.added_ids())
         }
 
         let maybe_wc_commit_id = self
@@ -905,37 +918,50 @@ impl MutableRepo {
     }
 
     pub fn add_head(&mut self, head: &Commit) {
+        self.add_heads(slice::from_ref(head));
+    }
+
+    pub fn add_heads(&mut self, heads: &[Commit]) {
         let current_heads = self.view.get_mut().heads();
         // Use incremental update for common case of adding a single commit on top a
         // current head. TODO: Also use incremental update when adding a single
         // commit on top a non-head.
-        if head
-            .parent_ids()
-            .iter()
-            .all(|parent_id| current_heads.contains(parent_id))
-        {
-            self.index.add_commit(head);
-            self.view.get_mut().add_head(head.id());
-            for parent_id in head.parent_ids() {
-                self.view.get_mut().remove_head(parent_id);
+        match heads {
+            [] => {}
+            [head]
+                if head
+                    .parent_ids()
+                    .iter()
+                    .all(|parent_id| current_heads.contains(parent_id)) =>
+            {
+                self.index.add_commit(head);
+                self.view.get_mut().add_head(head.id());
+                for parent_id in head.parent_ids() {
+                    self.view.get_mut().remove_head(parent_id);
+                }
             }
-        } else {
-            let missing_commits = dag_walk::topo_order_forward(
-                vec![head.clone()],
-                |commit: &Commit| commit.id().clone(),
-                |commit: &Commit| -> Vec<Commit> {
-                    commit
-                        .parents()
-                        .into_iter()
-                        .filter(|parent| !self.index().has_id(parent.id()))
-                        .collect()
-                },
-            );
-            for missing_commit in &missing_commits {
-                self.index.add_commit(missing_commit);
+            _ => {
+                let missing_commits = dag_walk::topo_order_reverse_ord(
+                    heads.iter().cloned().map(CommitByCommitterTimestamp),
+                    |CommitByCommitterTimestamp(commit)| commit.id().clone(),
+                    |CommitByCommitterTimestamp(commit)| {
+                        commit
+                            .parent_ids()
+                            .iter()
+                            .filter(|id| !self.index().has_id(id))
+                            .map(|id| self.store().get_commit(id).unwrap())
+                            .map(CommitByCommitterTimestamp)
+                            .collect_vec()
+                    },
+                );
+                for CommitByCommitterTimestamp(missing_commit) in missing_commits.iter().rev() {
+                    self.index.add_commit(missing_commit);
+                }
+                for head in heads {
+                    self.view.get_mut().add_head(head.id());
+                }
+                self.view.mark_dirty();
             }
-            self.view.get_mut().add_head(head.id());
-            self.view.mark_dirty();
         }
     }
 
@@ -966,25 +992,20 @@ impl MutableRepo {
         self.view_mut().remove_branch(name);
     }
 
-    pub fn get_local_branch(&self, name: &str) -> Option<RefTarget> {
-        self.view.with_ref(|v| v.get_local_branch(name))
+    pub fn get_local_branch(&self, name: &str) -> RefTarget {
+        self.view.with_ref(|v| v.get_local_branch(name).clone())
     }
 
-    pub fn set_local_branch_target(&mut self, name: &str, target: Option<RefTarget>) {
+    pub fn set_local_branch_target(&mut self, name: &str, target: RefTarget) {
         self.view_mut().set_local_branch_target(name, target);
     }
 
-    pub fn get_remote_branch(&self, name: &str, remote_name: &str) -> Option<RefTarget> {
+    pub fn get_remote_branch(&self, name: &str, remote_name: &str) -> RefTarget {
         self.view
-            .with_ref(|v| v.get_remote_branch(name, remote_name))
+            .with_ref(|v| v.get_remote_branch(name, remote_name).clone())
     }
 
-    pub fn set_remote_branch_target(
-        &mut self,
-        name: &str,
-        remote_name: &str,
-        target: Option<RefTarget>,
-    ) {
+    pub fn set_remote_branch_target(&mut self, name: &str, remote_name: &str, target: RefTarget) {
         self.view_mut()
             .set_remote_branch_target(name, remote_name, target);
     }
@@ -993,27 +1014,27 @@ impl MutableRepo {
         self.view_mut().rename_remote(old, new);
     }
 
-    pub fn get_tag(&self, name: &str) -> Option<RefTarget> {
-        self.view.with_ref(|v| v.get_tag(name))
+    pub fn get_tag(&self, name: &str) -> RefTarget {
+        self.view.with_ref(|v| v.get_tag(name).clone())
     }
 
-    pub fn set_tag_target(&mut self, name: &str, target: Option<RefTarget>) {
+    pub fn set_tag_target(&mut self, name: &str, target: RefTarget) {
         self.view_mut().set_tag_target(name, target);
     }
 
-    pub fn get_git_ref(&self, name: &str) -> Option<RefTarget> {
-        self.view.with_ref(|v| v.get_git_ref(name))
+    pub fn get_git_ref(&self, name: &str) -> RefTarget {
+        self.view.with_ref(|v| v.get_git_ref(name).clone())
     }
 
-    pub fn set_git_ref_target(&mut self, name: &str, target: Option<RefTarget>) {
+    pub fn set_git_ref_target(&mut self, name: &str, target: RefTarget) {
         self.view_mut().set_git_ref_target(name, target);
     }
 
-    pub fn git_head(&self) -> Option<RefTarget> {
-        self.view.with_ref(|v| v.git_head().cloned())
+    pub fn git_head(&self) -> RefTarget {
+        self.view.with_ref(|v| v.git_head().clone())
     }
 
-    pub fn set_git_head_target(&mut self, target: Option<RefTarget>) {
+    pub fn set_git_head_target(&mut self, target: RefTarget) {
         self.view_mut().set_git_head_target(target);
     }
 
@@ -1087,8 +1108,8 @@ impl MutableRepo {
         let base_branches: HashSet<_> = base.branches().keys().cloned().collect();
         let other_branches: HashSet<_> = other.branches().keys().cloned().collect();
         for branch_name in base_branches.union(&other_branches) {
-            let base_branch = base.branches().get(branch_name);
-            let other_branch = other.branches().get(branch_name);
+            let base_branch = base.get_branch(branch_name);
+            let other_branch = other.get_branch(branch_name);
             if other_branch == base_branch {
                 // Unchanged on other side
                 continue;
@@ -1133,8 +1154,8 @@ impl MutableRepo {
             self.view.get_mut().merge_single_ref(
                 self.index.as_index(),
                 &ref_name,
-                base_target.as_ref(),
-                other_target.as_ref(),
+                base_target,
+                other_target,
             );
         }
 
@@ -1198,8 +1219,8 @@ impl MutableRepo {
     pub fn merge_single_ref(
         &mut self,
         ref_name: &RefName,
-        base_target: Option<&RefTarget>,
-        other_target: Option<&RefTarget>,
+        base_target: &RefTarget,
+        other_target: &RefTarget,
     ) {
         self.view.get_mut().merge_single_ref(
             self.index.as_index(),
