@@ -18,33 +18,39 @@ use std::collections::{HashMap, HashSet};
 use std::sync::Arc;
 
 use itertools::{process_results, Itertools};
+use tracing::instrument;
 
 use crate::backend::{BackendError, CommitId, ObjectId};
 use crate::commit::Commit;
 use crate::dag_walk;
 use crate::index::Index;
+use crate::merged_tree::MergedTree;
 use crate::op_store::RefTarget;
 use crate::repo::{MutableRepo, Repo};
-use crate::repo_path::RepoPath;
 use crate::revset::{RevsetExpression, RevsetIteratorExt};
 use crate::settings::UserSettings;
 use crate::store::Store;
-use crate::tree::{merge_trees, Tree, TreeMergeError};
+use crate::tree::TreeMergeError;
 use crate::view::RefName;
 
-pub fn merge_commit_trees(repo: &dyn Repo, commits: &[Commit]) -> Result<Tree, TreeMergeError> {
+#[instrument(skip(repo))]
+pub fn merge_commit_trees(
+    repo: &dyn Repo,
+    commits: &[Commit],
+) -> Result<MergedTree, TreeMergeError> {
     merge_commit_trees_without_repo(repo.store(), repo.index(), commits)
 }
 
+#[instrument(skip(index))]
 pub fn merge_commit_trees_without_repo(
     store: &Arc<Store>,
     index: &dyn Index,
     commits: &[Commit],
-) -> Result<Tree, TreeMergeError> {
+) -> Result<MergedTree, TreeMergeError> {
     if commits.is_empty() {
-        Ok(store.get_tree(&RepoPath::root(), store.empty_tree_id())?)
+        Ok(store.get_root_tree(&store.empty_merged_tree_id())?)
     } else {
-        let mut new_tree = commits[0].tree();
+        let mut new_tree = commits[0].tree()?;
         let commit_ids = commits
             .iter()
             .map(|commit| commit.id().clone())
@@ -56,7 +62,8 @@ pub fn merge_commit_trees_without_repo(
                 .map(|id| store.get_commit(id))
                 .try_collect()?;
             let ancestor_tree = merge_commit_trees_without_repo(store, index, &ancestors)?;
-            new_tree = merge_trees(&new_tree, &ancestor_tree, &other_commit.tree())?;
+            let other_tree = other_commit.tree()?;
+            new_tree = new_tree.merge(&ancestor_tree, &other_tree)?;
         }
         Ok(new_tree)
     }
@@ -83,8 +90,9 @@ pub fn rebase_commit(
     } else {
         let old_base_tree = merge_commit_trees(mut_repo, &old_parents)?;
         let new_base_tree = merge_commit_trees(mut_repo, new_parents)?;
-        let merged_tree = merge_trees(&new_base_tree, &old_base_tree, &old_commit.tree())?;
-        merged_tree.id().clone()
+        let old_tree = old_commit.tree()?;
+        let merged_tree = new_base_tree.merge(&old_base_tree, &old_tree)?;
+        merged_tree.id()
     };
     let new_parent_ids = new_parents
         .iter()
@@ -93,7 +101,7 @@ pub fn rebase_commit(
     Ok(mut_repo
         .rewrite_commit(settings, old_commit)
         .set_parents(new_parent_ids)
-        .set_tree(new_tree_id)
+        .set_tree_id(new_tree_id)
         .write()?)
 }
 
@@ -105,14 +113,15 @@ pub fn back_out_commit(
 ) -> Result<Commit, TreeMergeError> {
     let old_base_tree = merge_commit_trees(mut_repo, &old_commit.parents())?;
     let new_base_tree = merge_commit_trees(mut_repo, new_parents)?;
-    let new_tree = merge_trees(&new_base_tree, &old_commit.tree(), &old_base_tree).unwrap();
+    let old_tree = old_commit.tree()?;
+    let new_tree = new_base_tree.merge(&old_tree, &old_base_tree)?;
     let new_parent_ids = new_parents
         .iter()
         .map(|commit| commit.id().clone())
         .collect();
     // TODO: i18n the description based on repo language
     Ok(mut_repo
-        .new_commit(settings, new_parent_ids, new_tree.id().clone())
+        .new_commit(settings, new_parent_ids, new_tree.id())
         .set_description(format!("backout of commit {}", &old_commit.id().hex()))
         .write()?)
 }
@@ -225,13 +234,11 @@ impl<'settings, 'repo> DescendantRebaser<'settings, 'repo> {
         // all branches each time we rebase a commit.
         let mut branches: HashMap<_, HashSet<_>> = HashMap::new();
         for (branch_name, branch_target) in mut_repo.view().branches() {
-            if let Some(local_target) = &branch_target.local_target {
-                for commit in local_target.adds() {
-                    branches
-                        .entry(commit.clone())
-                        .or_default()
-                        .insert(branch_name.clone());
-                }
+            for commit in branch_target.local_target.added_ids() {
+                branches
+                    .entry(commit.clone())
+                    .or_default()
+                    .insert(branch_name.clone());
             }
         }
 
@@ -278,10 +285,7 @@ impl<'settings, 'repo> DescendantRebaser<'settings, 'repo> {
         new_ids
     }
 
-    fn ref_target_update(
-        old_id: CommitId,
-        new_ids: Vec<CommitId>,
-    ) -> (Option<RefTarget>, Option<RefTarget>) {
+    fn ref_target_update(old_id: CommitId, new_ids: Vec<CommitId>) -> (RefTarget, RefTarget) {
         let old_ids = std::iter::repeat(old_id).take(new_ids.len());
         (
             RefTarget::from_legacy_form([], old_ids),
@@ -307,8 +311,8 @@ impl<'settings, 'repo> DescendantRebaser<'settings, 'repo> {
                         .or_default()
                         .insert(branch_name.clone());
                 }
-                let local_target = self.mut_repo.get_local_branch(branch_name).unwrap();
-                for old_add in local_target.adds() {
+                let local_target = self.mut_repo.get_local_branch(branch_name);
+                for old_add in local_target.added_ids() {
                     if *old_add == old_commit_id {
                         branch_updates.push(branch_name.clone());
                     }
@@ -319,8 +323,8 @@ impl<'settings, 'repo> DescendantRebaser<'settings, 'repo> {
             for branch_name in branch_updates {
                 self.mut_repo.merge_single_ref(
                     &RefName::LocalBranch(branch_name),
-                    old_target.as_ref(),
-                    new_target.as_ref(),
+                    &old_target,
+                    &new_target,
                 );
             }
         }
