@@ -16,33 +16,46 @@
 
 use std::any::Any;
 use std::collections::HashMap;
+use std::fmt::{Debug, Formatter};
 use std::io::Read;
 use std::sync::{Arc, RwLock};
 
+use crate::backend;
 use crate::backend::{
-    Backend, BackendResult, ChangeId, CommitId, ConflictId, FileId, SymlinkId, TreeId, TreeValue,
+    Backend, BackendResult, ChangeId, CommitId, ConflictId, FileId, MergedTreeId, SymlinkId,
+    TreeId, TreeValue,
 };
 use crate::commit::Commit;
+use crate::merge::Merge;
+use crate::merged_tree::MergedTree;
 use crate::repo_path::RepoPath;
 use crate::tree::Tree;
 use crate::tree_builder::TreeBuilder;
-use crate::{backend, conflicts};
 
 /// Wraps the low-level backend and makes it return more convenient types. Also
 /// adds caching.
-#[derive(Debug)]
 pub struct Store {
     backend: Box<dyn Backend>,
     commit_cache: RwLock<HashMap<CommitId, Arc<backend::Commit>>>,
     tree_cache: RwLock<HashMap<(RepoPath, TreeId), Arc<backend::Tree>>>,
+    use_tree_conflict_format: bool,
+}
+
+impl Debug for Store {
+    fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), std::fmt::Error> {
+        f.debug_struct("Store")
+            .field("backend", &self.backend)
+            .finish_non_exhaustive()
+    }
 }
 
 impl Store {
-    pub fn new(backend: Box<dyn Backend>) -> Arc<Self> {
+    pub fn new(backend: Box<dyn Backend>, use_tree_conflict_format: bool) -> Arc<Self> {
         Arc::new(Store {
             backend,
             commit_cache: Default::default(),
             tree_cache: Default::default(),
+            use_tree_conflict_format,
         })
     }
 
@@ -50,6 +63,11 @@ impl Store {
         self.backend.as_any()
     }
 
+    /// Whether new tree should be written using the tree-level format.
+    pub fn use_tree_conflict_format(&self) -> bool {
+        self.use_tree_conflict_format
+    }
+
     pub fn commit_id_length(&self) -> usize {
         self.backend.commit_id_length()
     }
@@ -62,6 +80,10 @@ impl Store {
         self.backend.empty_tree_id()
     }
 
+    pub fn empty_merged_tree_id(&self) -> MergedTreeId {
+        MergedTreeId::Legacy(self.backend.empty_tree_id().clone())
+    }
+
     pub fn root_commit_id(&self) -> &CommitId {
         self.backend.root_commit_id()
     }
@@ -124,6 +146,19 @@ impl Store {
         Ok(data)
     }
 
+    pub fn get_root_tree(self: &Arc<Self>, id: &MergedTreeId) -> BackendResult<MergedTree> {
+        match &id {
+            MergedTreeId::Legacy(id) => {
+                let tree = self.get_tree(&RepoPath::root(), id)?;
+                Ok(MergedTree::Legacy(tree))
+            }
+            MergedTreeId::Merge(ids) => {
+                let trees = ids.try_map(|id| self.get_tree(&RepoPath::root(), id))?;
+                Ok(MergedTree::Merge(trees))
+            }
+        }
+    }
+
     pub fn write_tree(
         self: &Arc<Self>,
         path: &RepoPath,
@@ -159,15 +194,15 @@ impl Store {
         &self,
         path: &RepoPath,
         id: &ConflictId,
-    ) -> BackendResult<conflicts::Conflict<Option<TreeValue>>> {
+    ) -> BackendResult<Merge<Option<TreeValue>>> {
         let backend_conflict = self.backend.read_conflict(path, id)?;
-        Ok(conflicts::Conflict::from_backend_conflict(backend_conflict))
+        Ok(Merge::from_backend_conflict(backend_conflict))
     }
 
     pub fn write_conflict(
         &self,
         path: &RepoPath,
-        contents: &conflicts::Conflict<Option<TreeValue>>,
+        contents: &Merge<Option<TreeValue>>,
     ) -> BackendResult<ConflictId> {
         self.backend
             .write_conflict(path, &contents.clone().into_backend_conflict())
