@@ -21,15 +21,15 @@ use std::sync::Arc;
 
 use itertools::Itertools;
 use thiserror::Error;
+use tracing::instrument;
 
 use crate::backend::{
     BackendError, ConflictId, FileId, ObjectId, TreeEntriesNonRecursiveIterator, TreeEntry, TreeId,
     TreeValue,
 };
-use crate::conflicts::Conflict;
 use crate::files::MergeResult;
 use crate::matchers::{EverythingMatcher, Matcher};
-use crate::merge::trivial_merge;
+use crate::merge::{trivial_merge, Merge};
 use crate::repo_path::{RepoPath, RepoPathComponent, RepoPathJoin};
 use crate::store::Store;
 use crate::{backend, files};
@@ -77,19 +77,6 @@ impl Hash for Tree {
     }
 }
 
-#[derive(Debug, PartialEq, Eq, Clone)]
-pub struct DiffSummary {
-    pub modified: Vec<RepoPath>,
-    pub added: Vec<RepoPath>,
-    pub removed: Vec<RepoPath>,
-}
-
-impl DiffSummary {
-    pub fn is_empty(&self) -> bool {
-        self.modified.is_empty() && self.added.is_empty() && self.removed.is_empty()
-    }
-}
-
 impl Tree {
     pub fn new(store: Arc<Store>, dir: RepoPath, id: TreeId, data: Arc<backend::Tree>) -> Self {
         Tree {
@@ -184,35 +171,6 @@ impl Tree {
         }
     }
 
-    pub fn diff<'matcher>(
-        &self,
-        other: &Tree,
-        matcher: &'matcher dyn Matcher,
-    ) -> TreeDiffIterator<'matcher> {
-        TreeDiffIterator::new(self.clone(), other.clone(), matcher)
-    }
-
-    pub fn diff_summary(&self, other: &Tree, matcher: &dyn Matcher) -> DiffSummary {
-        let mut modified = vec![];
-        let mut added = vec![];
-        let mut removed = vec![];
-        for (file, diff) in self.diff(other, matcher) {
-            match diff {
-                Diff::Modified(_, _) => modified.push(file.clone()),
-                Diff::Added(_) => added.push(file.clone()),
-                Diff::Removed(_) => removed.push(file.clone()),
-            }
-        }
-        modified.sort();
-        added.sort();
-        removed.sort();
-        DiffSummary {
-            modified,
-            added,
-            removed,
-        }
-    }
-
     pub fn conflicts_matching(&self, matcher: &dyn Matcher) -> Vec<(RepoPath, ConflictId)> {
         let mut conflicts = vec![];
         for (name, value) in self.entries_matching(matcher) {
@@ -223,6 +181,7 @@ impl Tree {
         conflicts
     }
 
+    #[instrument]
     pub fn conflicts(&self) -> Vec<(RepoPath, ConflictId)> {
         self.conflicts_matching(&EverythingMatcher)
     }
@@ -296,32 +255,6 @@ impl Iterator for TreeEntriesIterator<'_> {
     }
 }
 
-#[derive(Debug, PartialEq, Eq, Clone)]
-pub enum Diff<T> {
-    Modified(T, T),
-    Added(T),
-    Removed(T),
-}
-
-impl<T> Diff<T> {
-    pub fn from_options(left: Option<T>, right: Option<T>) -> Self {
-        match (left, right) {
-            (Some(left), Some(right)) => Diff::Modified(left, right),
-            (None, Some(right)) => Diff::Added(right),
-            (Some(left), None) => Diff::Removed(left),
-            (None, None) => panic!("left and right cannot both be None"),
-        }
-    }
-
-    pub fn into_options(self) -> (Option<T>, Option<T>) {
-        match self {
-            Diff::Modified(left, right) => (Some(left), Some(right)),
-            Diff::Added(right) => (None, Some(right)),
-            Diff::Removed(left) => (Some(left), None),
-        }
-    }
-}
-
 struct TreeEntryDiffIterator<'trees> {
     tree1: &'trees Tree,
     tree2: &'trees Tree,
@@ -358,128 +291,6 @@ impl<'trees> Iterator for TreeEntryDiffIterator<'trees> {
     }
 }
 
-pub struct TreeDiffIterator<'matcher> {
-    stack: Vec<TreeDiffItem>,
-    matcher: &'matcher dyn Matcher,
-}
-
-struct TreeDiffDirItem {
-    path: RepoPath,
-    // Iterator over the diffs between tree1 and tree2
-    entry_iterator: TreeEntryDiffIterator<'static>,
-    // On drop, tree1 and tree2 must outlive entry_iterator
-    tree1: Box<Tree>,
-    tree2: Box<Tree>,
-}
-
-enum TreeDiffItem {
-    Dir(TreeDiffDirItem),
-    // This is used for making sure that when a directory gets replaced by a file, we
-    // yield the value for the addition of the file after we yield the values
-    // for removing files in the directory.
-    File(RepoPath, Diff<TreeValue>),
-}
-
-impl<'matcher> TreeDiffIterator<'matcher> {
-    fn new(tree1: Tree, tree2: Tree, matcher: &'matcher dyn Matcher) -> Self {
-        let root_dir = RepoPath::root();
-        let mut stack = Vec::new();
-        if !matcher.visit(&root_dir).is_nothing() {
-            stack.push(TreeDiffItem::Dir(TreeDiffDirItem::new(
-                root_dir, tree1, tree2,
-            )));
-        };
-        Self { stack, matcher }
-    }
-}
-
-impl TreeDiffDirItem {
-    fn new(path: RepoPath, tree1: Tree, tree2: Tree) -> Self {
-        let tree1 = Box::new(tree1);
-        let tree2 = Box::new(tree2);
-        let iter: TreeEntryDiffIterator = TreeEntryDiffIterator::new(&tree1, &tree2);
-        let iter: TreeEntryDiffIterator<'static> = unsafe { std::mem::transmute(iter) };
-        Self {
-            path,
-            entry_iterator: iter,
-            tree1,
-            tree2,
-        }
-    }
-
-    fn subdir(
-        &self,
-        subdir_path: &RepoPath,
-        before: Option<&TreeValue>,
-        after: Option<&TreeValue>,
-    ) -> Self {
-        let before_tree = match before {
-            Some(TreeValue::Tree(id_before)) => self.tree1.known_sub_tree(subdir_path, id_before),
-            _ => Tree::null(self.tree1.store().clone(), subdir_path.clone()),
-        };
-        let after_tree = match after {
-            Some(TreeValue::Tree(id_after)) => self.tree2.known_sub_tree(subdir_path, id_after),
-            _ => Tree::null(self.tree2.store().clone(), subdir_path.clone()),
-        };
-        Self::new(subdir_path.clone(), before_tree, after_tree)
-    }
-}
-
-impl Iterator for TreeDiffIterator<'_> {
-    type Item = (RepoPath, Diff<TreeValue>);
-
-    fn next(&mut self) -> Option<Self::Item> {
-        while let Some(top) = self.stack.last_mut() {
-            let (dir, (name, before, after)) = match top {
-                TreeDiffItem::Dir(dir) => {
-                    if let Some(entry) = dir.entry_iterator.next() {
-                        (dir, entry)
-                    } else {
-                        self.stack.pop().unwrap();
-                        continue;
-                    }
-                }
-                TreeDiffItem::File(..) => {
-                    if let TreeDiffItem::File(name, diff) = self.stack.pop().unwrap() {
-                        return Some((name, diff));
-                    } else {
-                        unreachable!();
-                    }
-                }
-            };
-
-            let path = dir.path.join(name);
-            let tree_before = matches!(before, Some(TreeValue::Tree(_)));
-            let tree_after = matches!(after, Some(TreeValue::Tree(_)));
-            let post_subdir =
-                if (tree_before || tree_after) && !self.matcher.visit(&path).is_nothing() {
-                    let subdir = dir.subdir(&path, before, after);
-                    self.stack.push(TreeDiffItem::Dir(subdir));
-                    self.stack.len() - 1
-                } else {
-                    self.stack.len()
-                };
-            if self.matcher.matches(&path) {
-                if !tree_before && tree_after {
-                    if let Some(value_before) = before {
-                        return Some((path, Diff::Removed(value_before.clone())));
-                    }
-                } else if tree_before && !tree_after {
-                    if let Some(value_after) = after {
-                        self.stack.insert(
-                            post_subdir,
-                            TreeDiffItem::File(path, Diff::Added(value_after.clone())),
-                        );
-                    }
-                } else if !tree_before && !tree_after {
-                    return Some((path, Diff::from_options(before.cloned(), after.cloned())));
-                }
-            }
-        }
-        None
-    }
-}
-
 pub fn merge_trees(
     side1_tree: &Tree,
     base_tree: &Tree,
@@ -560,57 +371,57 @@ fn merge_tree_value(
             }
         }
         _ => {
-            // Start by creating a Conflict object. Conflicts can cleanly represent a single
+            // Start by creating a Merge object. Merges can cleanly represent a single
             // resolved state, the absence of a state, or a conflicted state.
-            let conflict = Conflict::new(
+            let conflict = Merge::new(
                 vec![maybe_base.cloned()],
                 vec![maybe_side1.cloned(), maybe_side2.cloned()],
             );
             let filename = dir.join(basename);
-            let conflict = simplify_conflict(store, &filename, conflict)?;
-            if let Some(value) = conflict.as_resolved() {
-                return Ok(value.clone());
-            }
-            if let Some(tree_value) = try_resolve_file_conflict(store, &filename, &conflict)? {
-                Some(tree_value)
-            } else {
-                let conflict_id = store.write_conflict(&filename, &conflict)?;
-                Some(TreeValue::Conflict(conflict_id))
+            let expanded = conflict.try_map(|term| match term {
+                Some(TreeValue::Conflict(id)) => store.read_conflict(&filename, id),
+                _ => Ok(Merge::resolved(term.clone())),
+            })?;
+            let merge = expanded.flatten().simplify();
+            match merge.into_resolved() {
+                Ok(value) => value,
+                Err(conflict) => {
+                    if let Some(tree_value) =
+                        try_resolve_file_conflict(store, &filename, &conflict)?
+                    {
+                        Some(tree_value)
+                    } else {
+                        let conflict_id = store.write_conflict(&filename, &conflict)?;
+                        Some(TreeValue::Conflict(conflict_id))
+                    }
+                }
             }
         }
     })
 }
 
-fn try_resolve_file_conflict(
+pub fn try_resolve_file_conflict(
     store: &Store,
     filename: &RepoPath,
-    conflict: &Conflict<Option<TreeValue>>,
+    conflict: &Merge<Option<TreeValue>>,
 ) -> Result<Option<TreeValue>, TreeMergeError> {
     // If there are any non-file or any missing parts in the conflict, we can't
     // merge it. We check early so we don't waste time reading file contents if
     // we can't merge them anyway. At the same time we determine whether the
     // resulting file should be executable.
-    // TODO: Change to let-else once our MSRV is above 1.65
-    let file_id_conflict = if let Some(conflict) = conflict.maybe_map(|term| match term {
+    let Some(file_id_conflict) = conflict.maybe_map(|term| match term {
         Some(TreeValue::File { id, executable: _ }) => Some(id),
         _ => None,
-    }) {
-        conflict
-    } else {
+    }) else {
         return Ok(None);
     };
-    // TODO: Change to let-else once our MSRV is above 1.65
-    let executable_conflict = if let Some(conflict) = conflict.maybe_map(|term| match term {
+    let Some(executable_conflict) = conflict.maybe_map(|term| match term {
         Some(TreeValue::File { id: _, executable }) => Some(executable),
         _ => None,
-    }) {
-        conflict
-    } else {
+    }) else {
         return Ok(None);
     };
-    let executable = if let Some(&executable) = executable_conflict.resolve_trivial() {
-        *executable
-    } else {
+    let Some(&&executable) = executable_conflict.resolve_trivial() else {
         // We're unable to determine whether the result should be executable
         return Ok(None);
     };
@@ -658,46 +469,3 @@ fn try_resolve_file_conflict(
         MergeResult::Conflict(_) => Ok(None),
     }
 }
-
-fn simplify_conflict(
-    store: &Store,
-    path: &RepoPath,
-    conflict: Conflict<Option<TreeValue>>,
-) -> Result<Conflict<Option<TreeValue>>, BackendError> {
-    // Important cases to simplify:
-    //
-    // D
-    // |
-    // B C
-    // |/
-    // A
-    //
-    // 1. rebase C to B, then back to A => there should be no conflict
-    // 2. rebase C to B, then to D => the conflict should not mention B
-    // 3. rebase B to C and D to B', then resolve the conflict in B' and rebase D'
-    // on top =>    the conflict should be between B'', B, and D; it should not
-    // mention the conflict in B'
-
-    // Case 1 above:
-    // After first rebase, the conflict is {+B-A+C}. After rebasing back,
-    // the unsimplified conflict is {+A-B+{+B-A+C}}. Since the
-    // inner conflict is positive, we can simply move it into the outer conflict. We
-    // thus get {+A-B+B-A+C}, which we can then simplify to just C (because {+C} ==
-    // C).
-    //
-    // Case 2 above:
-    // After first rebase, the conflict is {+B-A+C}. After rebasing to D,
-    // the unsimplified conflict is {+D-C+{+B-A+C}}. As in the
-    // previous case, the inner conflict can be moved into the outer one. We then
-    // get {+D-C+B-A+C}. That can be simplified to
-    // {+D+B-A}, which is the desired conflict.
-    //
-    // Case 3 above:
-    // TODO: describe this case
-
-    let expanded = conflict.try_map(|term| match term {
-        Some(TreeValue::Conflict(id)) => store.read_conflict(path, id),
-        _ => Ok(Conflict::resolved(term.clone())),
-    })?;
-    Ok(expanded.flatten().simplify())
-}
