@@ -22,7 +22,7 @@ use itertools::Itertools;
 use crate::backend::CommitId;
 use crate::index::Index;
 use crate::op_store;
-use crate::op_store::{BranchTarget, RefTarget, WorkspaceId};
+use crate::op_store::{BranchTarget, RefTarget, RefTargetOptionExt as _, WorkspaceId};
 use crate::refs::merge_ref_targets;
 
 #[derive(PartialEq, Eq, PartialOrd, Ord, Clone, Hash, Debug)]
@@ -98,8 +98,8 @@ impl View {
         &self.data.git_refs
     }
 
-    pub fn git_head(&self) -> Option<&RefTarget> {
-        self.data.git_head.as_ref()
+    pub fn git_head(&self) -> &RefTarget {
+        &self.data.git_head
     }
 
     pub fn set_wc_commit(&mut self, workspace_id: WorkspaceId, commit_id: CommitId) {
@@ -126,7 +126,7 @@ impl View {
         self.data.public_head_ids.remove(head_id);
     }
 
-    pub fn get_ref(&self, name: &RefName) -> Option<RefTarget> {
+    pub fn get_ref(&self, name: &RefName) -> &RefTarget {
         match &name {
             RefName::LocalBranch(name) => self.get_local_branch(name),
             RefName::RemoteBranch { branch, remote } => self.get_remote_branch(branch, remote),
@@ -137,7 +137,7 @@ impl View {
 
     /// Sets reference of the specified kind to point to the given target. If
     /// the target is absent, the reference will be removed.
-    pub fn set_ref_target(&mut self, name: &RefName, target: Option<RefTarget>) {
+    pub fn set_ref_target(&mut self, name: &RefName, target: RefTarget) {
         match name {
             RefName::LocalBranch(name) => self.set_local_branch_target(name, target),
             RefName::RemoteBranch { branch, remote } => {
@@ -160,17 +160,18 @@ impl View {
         self.data.branches.remove(name);
     }
 
-    pub fn get_local_branch(&self, name: &str) -> Option<RefTarget> {
-        self.data
-            .branches
-            .get(name)
-            .and_then(|branch_target| branch_target.local_target.clone())
+    pub fn get_local_branch(&self, name: &str) -> &RefTarget {
+        if let Some(branch_target) = self.data.branches.get(name) {
+            &branch_target.local_target
+        } else {
+            RefTarget::absent_ref()
+        }
     }
 
     /// Sets local branch to point to the given target. If the target is absent,
     /// and if no associated remote branches exist, the branch will be removed.
-    pub fn set_local_branch_target(&mut self, name: &str, target: Option<RefTarget>) {
-        if let Some(target) = target {
+    pub fn set_local_branch_target(&mut self, name: &str, target: RefTarget) {
+        if target.is_present() {
             self.insert_local_branch(name.to_owned(), target);
         } else {
             self.remove_local_branch(name);
@@ -178,34 +179,32 @@ impl View {
     }
 
     fn insert_local_branch(&mut self, name: String, target: RefTarget) {
-        self.data.branches.entry(name).or_default().local_target = Some(target);
+        assert!(target.is_present());
+        self.data.branches.entry(name).or_default().local_target = target;
     }
 
     fn remove_local_branch(&mut self, name: &str) {
         if let Some(branch) = self.data.branches.get_mut(name) {
-            branch.local_target = None;
+            branch.local_target = RefTarget::absent();
             if branch.remote_targets.is_empty() {
                 self.remove_branch(name);
             }
         }
     }
 
-    pub fn get_remote_branch(&self, name: &str, remote_name: &str) -> Option<RefTarget> {
-        self.data
-            .branches
-            .get(name)
-            .and_then(|branch_target| branch_target.remote_targets.get(remote_name).cloned())
+    pub fn get_remote_branch(&self, name: &str, remote_name: &str) -> &RefTarget {
+        if let Some(branch_target) = self.data.branches.get(name) {
+            let maybe_target = branch_target.remote_targets.get(remote_name);
+            maybe_target.flatten()
+        } else {
+            RefTarget::absent_ref()
+        }
     }
 
     /// Sets remote-tracking branch to point to the given target. If the target
     /// is absent, the branch will be removed.
-    pub fn set_remote_branch_target(
-        &mut self,
-        name: &str,
-        remote_name: &str,
-        target: Option<RefTarget>,
-    ) {
-        if let Some(target) = target {
+    pub fn set_remote_branch_target(&mut self, name: &str, remote_name: &str, target: RefTarget) {
+        if target.is_present() {
             self.insert_remote_branch(name.to_owned(), remote_name.to_owned(), target);
         } else {
             self.remove_remote_branch(name, remote_name);
@@ -213,6 +212,7 @@ impl View {
     }
 
     fn insert_remote_branch(&mut self, name: String, remote_name: String, target: RefTarget) {
+        assert!(target.is_present());
         self.data
             .branches
             .entry(name)
@@ -224,7 +224,7 @@ impl View {
     fn remove_remote_branch(&mut self, name: &str, remote_name: &str) {
         if let Some(branch) = self.data.branches.get_mut(name) {
             branch.remote_targets.remove(remote_name);
-            if branch.remote_targets.is_empty() && branch.local_target.is_none() {
+            if branch.remote_targets.is_empty() && branch.local_target.is_absent() {
                 self.remove_branch(name);
             }
         }
@@ -232,34 +232,35 @@ impl View {
 
     pub fn rename_remote(&mut self, old: &str, new: &str) {
         for branch in self.data.branches.values_mut() {
-            if let Some(target) = branch.remote_targets.remove(old) {
+            let target = branch.remote_targets.remove(old).flatten();
+            if target.is_present() {
                 branch.remote_targets.insert(new.to_owned(), target);
             }
         }
     }
 
-    pub fn get_tag(&self, name: &str) -> Option<RefTarget> {
-        self.data.tags.get(name).cloned()
+    pub fn get_tag(&self, name: &str) -> &RefTarget {
+        self.data.tags.get(name).flatten()
     }
 
     /// Sets tag to point to the given target. If the target is absent, the tag
     /// will be removed.
-    pub fn set_tag_target(&mut self, name: &str, target: Option<RefTarget>) {
-        if let Some(target) = target {
+    pub fn set_tag_target(&mut self, name: &str, target: RefTarget) {
+        if target.is_present() {
             self.data.tags.insert(name.to_owned(), target);
         } else {
             self.data.tags.remove(name);
         }
     }
 
-    pub fn get_git_ref(&self, name: &str) -> Option<RefTarget> {
-        self.data.git_refs.get(name).cloned()
+    pub fn get_git_ref(&self, name: &str) -> &RefTarget {
+        self.data.git_refs.get(name).flatten()
     }
 
     /// Sets the last imported Git ref to point to the given target. If the
     /// target is absent, the reference will be removed.
-    pub fn set_git_ref_target(&mut self, name: &str, target: Option<RefTarget>) {
-        if let Some(target) = target {
+    pub fn set_git_ref_target(&mut self, name: &str, target: RefTarget) {
+        if target.is_present() {
             self.data.git_refs.insert(name.to_owned(), target);
         } else {
             self.data.git_refs.remove(name);
@@ -268,7 +269,7 @@ impl View {
 
     /// Sets `HEAD@git` to point to the given target. If the target is absent,
     /// the reference will be cleared.
-    pub fn set_git_head_target(&mut self, target: Option<RefTarget>) {
+    pub fn set_git_head_target(&mut self, target: RefTarget) {
         self.data.git_head = target;
     }
 
@@ -288,14 +289,13 @@ impl View {
         &mut self,
         index: &dyn Index,
         ref_name: &RefName,
-        base_target: Option<&RefTarget>,
-        other_target: Option<&RefTarget>,
+        base_target: &RefTarget,
+        other_target: &RefTarget,
     ) {
         if base_target != other_target {
             let self_target = self.get_ref(ref_name);
-            let new_target =
-                merge_ref_targets(index, self_target.as_ref(), base_target, other_target);
-            if new_target != self_target {
+            let new_target = merge_ref_targets(index, self_target, base_target, other_target);
+            if new_target != *self_target {
                 self.set_ref_target(ref_name, new_target);
             }
         }
