@@ -14,12 +14,12 @@
 
 use jj_lib::backend::{ChangeId, MillisSinceEpoch, ObjectId, Signature, Timestamp};
 use jj_lib::matchers::EverythingMatcher;
+use jj_lib::merged_tree::DiffSummary;
 use jj_lib::repo::Repo;
 use jj_lib::repo_path::RepoPath;
 use jj_lib::settings::UserSettings;
-use jj_lib::tree::DiffSummary;
 use test_case::test_case;
-use testutils::{assert_rebased, CommitGraphBuilder, TestRepo};
+use testutils::{assert_rebased, create_tree, CommitGraphBuilder, TestRepo};
 
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
@@ -31,7 +31,7 @@ fn test_initial(use_git: bool) {
 
     let root_file_path = RepoPath::from_internal_string("file");
     let dir_file_path = RepoPath::from_internal_string("dir/file");
-    let tree = testutils::create_tree(
+    let tree = create_tree(
         repo,
         &[
             (&root_file_path, "file contents"),
@@ -59,18 +59,14 @@ fn test_initial(use_git: bool) {
     let change_id = ChangeId::new(vec![100u8; 16]);
     let builder = tx
         .mut_repo()
-        .new_commit(
-            &settings,
-            vec![store.root_commit_id().clone()],
-            tree.id().clone(),
-        )
+        .new_commit(&settings, vec![store.root_commit_id().clone()], tree.id())
         .set_change_id(change_id.clone())
         .set_description("description")
         .set_author(author_signature.clone())
         .set_committer(committer_signature.clone());
     assert_eq!(builder.parents(), &[store.root_commit_id().clone()]);
     assert_eq!(builder.predecessors(), &[]);
-    assert_eq!(builder.tree(), tree.id());
+    assert_eq!(builder.tree_id(), &tree.id());
     assert_eq!(builder.change_id(), &change_id);
     assert_eq!(builder.author(), &author_signature);
     assert_eq!(builder.committer(), &committer_signature);
@@ -86,7 +82,8 @@ fn test_initial(use_git: bool) {
         store
             .root_commit()
             .tree()
-            .diff_summary(&commit.tree(), &EverythingMatcher),
+            .unwrap()
+            .diff_summary(&commit.tree().unwrap(), &EverythingMatcher),
         DiffSummary {
             modified: vec![],
             added: vec![dir_file_path, root_file_path],
@@ -105,7 +102,7 @@ fn test_rewrite(use_git: bool) {
 
     let root_file_path = RepoPath::from_internal_string("file");
     let dir_file_path = RepoPath::from_internal_string("dir/file");
-    let initial_tree = testutils::create_tree(
+    let initial_tree = create_tree(
         repo,
         &[
             (&root_file_path, "file contents"),
@@ -119,13 +116,13 @@ fn test_rewrite(use_git: bool) {
         .new_commit(
             &settings,
             vec![store.root_commit_id().clone()],
-            initial_tree.id().clone(),
+            initial_tree.id(),
         )
         .write()
         .unwrap();
     let repo = tx.commit();
 
-    let rewritten_tree = testutils::create_tree(
+    let rewritten_tree = create_tree(
         &repo,
         &[
             (&root_file_path, "file contents"),
@@ -145,7 +142,7 @@ fn test_rewrite(use_git: bool) {
     let rewritten_commit = tx
         .mut_repo()
         .rewrite_commit(&rewrite_settings, &initial_commit)
-        .set_tree(rewritten_tree.id().clone())
+        .set_tree_id(rewritten_tree.id().clone())
         .write()
         .unwrap();
     tx.mut_repo().rebase_descendants(&settings).unwrap();
@@ -169,7 +166,8 @@ fn test_rewrite(use_git: bool) {
         store
             .root_commit()
             .tree()
-            .diff_summary(&rewritten_commit.tree(), &EverythingMatcher),
+            .unwrap()
+            .diff_summary(&rewritten_commit.tree().unwrap(), &EverythingMatcher),
         DiffSummary {
             modified: vec![],
             added: vec![dir_file_path.clone(), root_file_path],
@@ -179,7 +177,8 @@ fn test_rewrite(use_git: bool) {
     assert_eq!(
         initial_commit
             .tree()
-            .diff_summary(&rewritten_commit.tree(), &EverythingMatcher),
+            .unwrap()
+            .diff_summary(&rewritten_commit.tree().unwrap(), &EverythingMatcher),
         DiffSummary {
             modified: vec![dir_file_path],
             added: vec![],
@@ -188,8 +187,7 @@ fn test_rewrite(use_git: bool) {
     );
 }
 
-// An author field with the placeholder name/email should get filled in on
-// rewrite
+// An author field with an empty name/email should get filled in on rewrite
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
 fn test_rewrite_update_missing_user(use_git: bool) {
@@ -204,14 +202,14 @@ fn test_rewrite_update_missing_user(use_git: bool) {
         .new_commit(
             &missing_user_settings,
             vec![repo.store().root_commit_id().clone()],
-            repo.store().empty_tree_id().clone(),
+            repo.store().empty_merged_tree_id(),
         )
         .write()
         .unwrap();
-    assert_eq!(initial_commit.author().name, "(no name configured)");
-    assert_eq!(initial_commit.author().email, "(no email configured)");
-    assert_eq!(initial_commit.committer().name, "(no name configured)");
-    assert_eq!(initial_commit.committer().email, "(no email configured)");
+    assert_eq!(initial_commit.author().name, "");
+    assert_eq!(initial_commit.author().email, "");
+    assert_eq!(initial_commit.committer().name, "");
+    assert_eq!(initial_commit.committer().email, "");
 
     let config = config::Config::builder()
         .set_override("user.name", "Configured User")
@@ -260,7 +258,7 @@ fn test_commit_builder_descendants(use_git: bool) {
         .new_commit(
             &settings,
             vec![store.root_commit_id().clone()],
-            store.empty_tree_id().clone(),
+            store.empty_merged_tree_id(),
         )
         .write()
         .unwrap();
