@@ -12,20 +12,16 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use jj_lib::backend::{FileId, TreeValue};
-use jj_lib::conflicts::{parse_conflict, Conflict};
+use jj_lib::backend::FileId;
+use jj_lib::conflicts::{
+    extract_as_single_hunk, materialize_merge_result, parse_conflict, update_from_content,
+};
+use jj_lib::merge::Merge;
 use jj_lib::repo::Repo;
 use jj_lib::repo_path::RepoPath;
 use jj_lib::store::Store;
 use testutils::TestRepo;
 
-fn file_value(file_id: &FileId) -> TreeValue {
-    TreeValue::File {
-        id: file_id.clone(),
-        executable: false,
-    }
-}
-
 #[test]
 fn test_materialize_conflict_basic() {
     let test_repo = TestRepo::init(false);
@@ -67,9 +63,9 @@ line 5
 
     // The left side should come first. The diff should be use the smaller (right)
     // side, and the left side should be a snapshot.
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![Some(file_value(&left_id)), Some(file_value(&right_id))],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(left_id.clone()), Some(right_id.clone())],
     );
     insta::assert_snapshot!(
         &materialize_conflict_string(store, &path, &conflict),
@@ -91,9 +87,9 @@ line 5
     );
     // Swap the positive terms in the conflict. The diff should still use the right
     // side, but now the right side should come first.
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![Some(file_value(&right_id)), Some(file_value(&left_id))],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(right_id.clone()), Some(left_id.clone())],
     );
     insta::assert_snapshot!(
         &materialize_conflict_string(store, &path, &conflict),
@@ -160,13 +156,9 @@ line 3
 
     // The order of (a, b, c) should be preserved. For all cases, the "a" side
     // should be a snapshot.
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id)), Some(file_value(&base_id))],
-        vec![
-            Some(file_value(&a_id)),
-            Some(file_value(&b_id)),
-            Some(file_value(&c_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone()), Some(base_id.clone())],
+        vec![Some(a_id.clone()), Some(b_id.clone()), Some(c_id.clone())],
     );
     insta::assert_snapshot!(
         &materialize_conflict_string(store, &path, &conflict),
@@ -188,13 +180,9 @@ line 3
     line 3
     "###
     );
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id)), Some(file_value(&base_id))],
-        vec![
-            Some(file_value(&c_id)),
-            Some(file_value(&b_id)),
-            Some(file_value(&a_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone()), Some(base_id.clone())],
+        vec![Some(c_id.clone()), Some(b_id.clone()), Some(a_id.clone())],
     );
     insta::assert_snapshot!(
         &materialize_conflict_string(store, &path, &conflict),
@@ -216,13 +204,9 @@ line 3
     line 3
     "###
     );
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id)), Some(file_value(&base_id))],
-        vec![
-            Some(file_value(&c_id)),
-            Some(file_value(&a_id)),
-            Some(file_value(&b_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone()), Some(base_id.clone())],
+        vec![Some(c_id.clone()), Some(a_id.clone()), Some(b_id.clone())],
     );
     insta::assert_snapshot!(
         &materialize_conflict_string(store, &path, &conflict),
@@ -283,14 +267,13 @@ line 5 right
 ",
     );
 
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![Some(file_value(&left_id)), Some(file_value(&right_id))],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(left_id.clone()), Some(right_id.clone())],
     );
-    let mut result: Vec<u8> = vec![];
-    conflict.materialize(store, &path, &mut result).unwrap();
+    let materialized = materialize_conflict_string(store, &path, &conflict);
     insta::assert_snapshot!(
-        String::from_utf8(result.clone()).unwrap(),
+        materialized,
         @r###"
     <<<<<<<
     +++++++
@@ -316,11 +299,11 @@ line 5 right
 
     // The first add should always be from the left side
     insta::assert_debug_snapshot!(
-        parse_conflict(&result, conflict.removes().len(), conflict.adds().len()),
+        parse_conflict(materialized.as_bytes(), conflict.num_sides()),
         @r###"
     Some(
         [
-            Conflict {
+            Merge {
                 removes: [
                     "line 1\nline 2\n",
                 ],
@@ -329,13 +312,13 @@ line 5 right
                     "line 1 right\nline 2\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [],
                 adds: [
                     "line 3\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [
                     "line 4\nline 5\n",
                 ],
@@ -386,12 +369,9 @@ line 5
     );
 
     // left modifies a line, right deletes the same line.
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![
-            Some(file_value(&modified_id)),
-            Some(file_value(&deleted_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(modified_id.clone()), Some(deleted_id.clone())],
     );
     insta::assert_snapshot!(&materialize_conflict_string(store, &path, &conflict), @r###"
     line 1
@@ -408,12 +388,9 @@ line 5
     );
 
     // right modifies a line, left deletes the same line.
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![
-            Some(file_value(&deleted_id)),
-            Some(file_value(&modified_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(deleted_id.clone()), Some(modified_id.clone())],
     );
     insta::assert_snapshot!(&materialize_conflict_string(store, &path, &conflict), @r###"
     line 1
@@ -430,9 +407,9 @@ line 5
     );
 
     // modify/delete conflict at the file level
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_id))],
-        vec![Some(file_value(&modified_id)), None],
+    let conflict = Merge::new(
+        vec![Some(base_id.clone())],
+        vec![Some(modified_id.clone()), None],
     );
     insta::assert_snapshot!(&materialize_conflict_string(store, &path, &conflict), @r###"
     <<<<<<<
@@ -459,7 +436,6 @@ line 3
 line 4
 line 5
 ",
-            1,
             2
         ),
         None
@@ -482,19 +458,18 @@ right
 >>>>>>>
 line 5
 ",
-            1,
             2
         ),
         @r###"
     Some(
         [
-            Conflict {
+            Merge {
                 removes: [],
                 adds: [
                     "line 1\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [
                     "line 2\nline 3\nline 4\n",
                 ],
@@ -503,7 +478,7 @@ line 5
                     "right\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [],
                 adds: [
                     "line 5\n",
@@ -536,19 +511,18 @@ right
 >>>>>>>
 line 5
 ",
-            2,
             3
         ),
         @r###"
     Some(
         [
-            Conflict {
+            Merge {
                 removes: [],
                 adds: [
                     "line 1\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [
                     "line 2\nline 3\nline 4\n",
                     "line 2\nline 3\nline 4\n",
@@ -559,7 +533,7 @@ line 5
                     "line 2\nforward\nline 3\nline 4\n",
                 ],
             },
-            Conflict {
+            Merge {
                 removes: [],
                 adds: [
                     "line 5\n",
@@ -587,7 +561,6 @@ right
 >>>>>>>
 line 5
 ",
-            2,
             3
         ),
         None
@@ -610,7 +583,6 @@ right
 >>>>>>>
 line 5
 ",
-            1,
             2
         ),
         None
@@ -634,7 +606,6 @@ right
 >>>>>>>
 line 5
 ",
-            1,
             2
         ),
         None
@@ -650,40 +621,31 @@ fn test_update_conflict_from_content() {
     let base_file_id = testutils::write_file(store, &path, "line 1\nline 2\nline 3\n");
     let left_file_id = testutils::write_file(store, &path, "left 1\nline 2\nleft 3\n");
     let right_file_id = testutils::write_file(store, &path, "right 1\nline 2\nright 3\n");
-    let conflict = Conflict::new(
-        vec![Some(file_value(&base_file_id))],
-        vec![
-            Some(file_value(&left_file_id)),
-            Some(file_value(&right_file_id)),
-        ],
+    let conflict = Merge::new(
+        vec![Some(base_file_id.clone())],
+        vec![Some(left_file_id.clone()), Some(right_file_id.clone())],
     );
 
     // If the content is unchanged compared to the materialized value, we get the
     // old conflict id back.
-    let mut materialized = vec![];
-    conflict
-        .materialize(store, &path, &mut materialized)
-        .unwrap();
-    let result = conflict
-        .update_from_content(store, &path, &materialized)
-        .unwrap();
-    assert_eq!(result, Some(conflict.clone()));
+    let materialized = materialize_conflict_string(store, &path, &conflict);
+    let result = update_from_content(&conflict, store, &path, materialized.as_bytes()).unwrap();
+    assert_eq!(result, conflict);
 
     // If the conflict is resolved, we get None back to indicate that.
-    let result = conflict
-        .update_from_content(store, &path, b"resolved 1\nline 2\nresolved 3\n")
-        .unwrap();
-    assert_eq!(result, None);
+    let result =
+        update_from_content(&conflict, store, &path, b"resolved 1\nline 2\nresolved 3\n").unwrap();
+    let expected_file_id = testutils::write_file(store, &path, "resolved 1\nline 2\nresolved 3\n");
+    assert_eq!(result, Merge::normal(expected_file_id));
 
     // If the conflict is partially resolved, we get a new conflict back.
-    let result = conflict
-        .update_from_content(
-            store,
-            &path,
-            b"resolved 1\nline 2\n<<<<<<<\n%%%%%%%\n-line 3\n+left 3\n+++++++\nright 3\n>>>>>>>\n",
-        )
-        .unwrap();
-    let new_conflict = result.unwrap();
+    let new_conflict = update_from_content(
+        &conflict,
+        store,
+        &path,
+        b"resolved 1\nline 2\n<<<<<<<\n%%%%%%%\n-line 3\n+left 3\n+++++++\nright 3\n>>>>>>>\n",
+    )
+    .unwrap();
     assert_ne!(new_conflict, conflict);
     // Calculate expected new FileIds
     let new_base_file_id = testutils::write_file(store, &path, "resolved 1\nline 2\nline 3\n");
@@ -691,11 +653,11 @@ fn test_update_conflict_from_content() {
     let new_right_file_id = testutils::write_file(store, &path, "resolved 1\nline 2\nright 3\n");
     assert_eq!(
         new_conflict,
-        Conflict::new(
-            vec![Some(file_value(&new_base_file_id))],
+        Merge::new(
+            vec![Some(new_base_file_id.clone())],
             vec![
-                Some(file_value(&new_left_file_id)),
-                Some(file_value(&new_right_file_id))
+                Some(new_left_file_id.clone()),
+                Some(new_right_file_id.clone())
             ]
         )
     );
@@ -709,36 +671,26 @@ fn test_update_conflict_from_content_modify_delete() {
     let path = RepoPath::from_internal_string("dir/file");
     let before_file_id = testutils::write_file(store, &path, "line 1\nline 2 before\nline 3\n");
     let after_file_id = testutils::write_file(store, &path, "line 1\nline 2 after\nline 3\n");
-    let conflict = Conflict::new(
-        vec![Some(file_value(&before_file_id))],
-        vec![Some(file_value(&after_file_id)), None],
-    );
+    let conflict = Merge::new(vec![Some(before_file_id)], vec![Some(after_file_id), None]);
 
     // If the content is unchanged compared to the materialized value, we get the
     // old conflict id back.
-    let mut materialized = vec![];
-    conflict
-        .materialize(store, &path, &mut materialized)
-        .unwrap();
-    let result = conflict
-        .update_from_content(store, &path, &materialized)
-        .unwrap();
-    assert_eq!(result, Some(conflict.clone()));
+    let materialized = materialize_conflict_string(store, &path, &conflict);
+    let result = update_from_content(&conflict, store, &path, materialized.as_bytes()).unwrap();
+    assert_eq!(result, conflict);
 
     // If the conflict is resolved, we get None back to indicate that.
-    let result = conflict
-        .update_from_content(store, &path, b"resolved\n")
-        .unwrap();
-    assert_eq!(result, None);
+    let result = update_from_content(&conflict, store, &path, b"resolved\n").unwrap();
+    let expected_file_id = testutils::write_file(store, &path, "resolved\n");
+    assert_eq!(result, Merge::normal(expected_file_id));
 
     // If the conflict is modified, we get a new conflict back.
-    let result = conflict.update_from_content(
+    let new_conflict = update_from_content(&conflict,
         store,
         &path,
         b"<<<<<<<\n%%%%%%%\n line 1\n-line 2 before\n+line 2 modified after\n line 3\n+++++++\n>>>>>>>\n",
     )
     .unwrap();
-    let new_conflict = result.unwrap();
     // Calculate expected new FileIds
     let new_base_file_id = testutils::write_file(store, &path, "line 1\nline 2 before\nline 3\n");
     let new_left_file_id =
@@ -746,9 +698,9 @@ fn test_update_conflict_from_content_modify_delete() {
 
     assert_eq!(
         new_conflict,
-        Conflict::new(
-            vec![Some(file_value(&new_base_file_id))],
-            vec![Some(file_value(&new_left_file_id)), None]
+        Merge::new(
+            vec![Some(new_base_file_id.clone())],
+            vec![Some(new_left_file_id.clone()), None]
         )
     );
 }
@@ -756,9 +708,10 @@ fn test_update_conflict_from_content_modify_delete() {
 fn materialize_conflict_string(
     store: &Store,
     path: &RepoPath,
-    conflict: &Conflict<Option<TreeValue>>,
+    conflict: &Merge<Option<FileId>>,
 ) -> String {
     let mut result: Vec<u8> = vec![];
-    conflict.materialize(store, path, &mut result).unwrap();
+    let contents = extract_as_single_hunk(conflict, store, path);
+    materialize_merge_result(&contents, &mut result).unwrap();
     String::from_utf8(result).unwrap()
 }
