@@ -13,10 +13,10 @@
 // limitations under the License.
 
 use jj_lib::matchers::{EverythingMatcher, FilesMatcher};
+use jj_lib::merged_tree::DiffSummary;
 use jj_lib::repo_path::RepoPath;
-use jj_lib::tree::DiffSummary;
 use test_case::test_case;
-use testutils::TestRepo;
+use testutils::{create_tree, TestRepo};
 
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
@@ -29,7 +29,7 @@ fn test_types(use_git: bool) {
     let added_path = RepoPath::from_internal_string("added");
     let removed_path = RepoPath::from_internal_string("removed");
 
-    let tree1 = testutils::create_tree(
+    let tree1 = create_tree(
         repo,
         &[
             (&clean_path, "clean"),
@@ -38,7 +38,7 @@ fn test_types(use_git: bool) {
         ],
     );
 
-    let tree2 = testutils::create_tree(
+    let tree2 = create_tree(
         repo,
         &[
             (&clean_path, "clean"),
@@ -66,8 +66,8 @@ fn test_tree_file_transition(use_git: bool) {
     let dir_file_path = RepoPath::from_internal_string("dir/file");
     let dir_path = RepoPath::from_internal_string("dir");
 
-    let tree1 = testutils::create_tree(repo, &[(&dir_file_path, "contents")]);
-    let tree2 = testutils::create_tree(repo, &[(&dir_path, "contents")]);
+    let tree1 = create_tree(repo, &[(&dir_file_path, "contents")]);
+    let tree2 = create_tree(repo, &[(&dir_path, "contents")]);
 
     assert_eq!(
         tree1.diff_summary(&tree2, &EverythingMatcher),
@@ -103,7 +103,7 @@ fn test_sorting(use_git: bool) {
     let s_b_path = RepoPath::from_internal_string("s/b");
     let z_path = RepoPath::from_internal_string("z");
 
-    let tree1 = testutils::create_tree(
+    let tree1 = create_tree(
         repo,
         &[
             (&a_path, "before"),
@@ -112,7 +112,7 @@ fn test_sorting(use_git: bool) {
         ],
     );
 
-    let tree2 = testutils::create_tree(
+    let tree2 = create_tree(
         repo,
         &[
             (&a_path, "after"),
@@ -161,8 +161,8 @@ fn test_matcher_dir_file_transition(use_git: bool) {
     let a_path = RepoPath::from_internal_string("a");
     let a_a_path = RepoPath::from_internal_string("a/a");
 
-    let tree1 = testutils::create_tree(repo, &[(&a_path, "before")]);
-    let tree2 = testutils::create_tree(repo, &[(&a_a_path, "after")]);
+    let tree1 = create_tree(repo, &[(&a_path, "before")]);
+    let tree2 = create_tree(repo, &[(&a_a_path, "after")]);
 
     let matcher = FilesMatcher::new(&[a_path.clone()]);
     assert_eq!(
@@ -230,12 +230,12 @@ fn test_matcher_normal_cases(use_git: bool) {
     let dir2_b_path = RepoPath::from_internal_string("dir2/b");
     let z_path = RepoPath::from_internal_string("z");
 
-    let tree1 = testutils::create_tree(repo, &[(&a_path, "before"), (&dir1_a_path, "before")]);
+    let tree1 = create_tree(repo, &[(&a_path, "before"), (&dir1_a_path, "before")]);
     // File "a" gets modified
     // File "dir1/a" gets modified
     // File "dir2/b" gets created
     // File "z" gets created
-    let tree2 = testutils::create_tree(
+    let tree2 = create_tree(
         repo,
         &[
             (&a_path, "after"),
