@@ -12,7 +12,6 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use assert_matches::assert_matches;
 use itertools::Itertools;
 use jj_lib::backend::TreeValue;
 use jj_lib::repo::Repo;
@@ -20,7 +19,7 @@ use jj_lib::repo_path::{RepoPath, RepoPathComponent};
 use jj_lib::rewrite::rebase_commit;
 use jj_lib::tree::{merge_trees, Tree};
 use test_case::test_case;
-use testutils::TestRepo;
+use testutils::{create_single_tree, create_tree, TestRepo};
 
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
@@ -54,7 +53,7 @@ fn test_same_type(use_git: bool) {
     let write_tree = |index: usize| -> Tree {
         let mut tree_builder = store.tree_builder(store.empty_tree_id().clone());
         for path in &files {
-            let contents = &path[index..index + 1];
+            let contents = &path[index..][..1];
             if contents != "_" {
                 testutils::write_normal_file(
                     &mut tree_builder,
@@ -216,7 +215,7 @@ fn test_executable(use_git: bool) {
     fn contents_in_tree<'a>(files: &[&'a str], index: usize) -> Vec<(&'a str, bool)> {
         files
             .iter()
-            .map(|f| (*f, &f[index..index + 1] == "x"))
+            .map(|f| (*f, &f[index..][..1] == "x"))
             .collect()
     }
 
@@ -476,8 +475,7 @@ fn test_simplify_conflict(use_git: bool) {
 
     let component = RepoPathComponent::from("file");
     let path = RepoPath::from_internal_string("file");
-    let write_tree =
-        |contents: &str| -> Tree { testutils::create_tree(repo, &[(&path, contents)]) };
+    let write_tree = |contents: &str| -> Tree { create_single_tree(repo, &[(&path, contents)]) };
 
     let base_tree = write_tree("base contents");
     let branch_tree = write_tree("branch contents");
@@ -572,32 +570,32 @@ fn test_simplify_conflict_after_resolving_parent(use_git: bool) {
     // a conflict since it changed an unrelated line.
     let path = RepoPath::from_internal_string("dir/file");
     let mut tx = repo.start_transaction(&settings, "test");
-    let tree_a = testutils::create_tree(repo, &[(&path, "abc\ndef\nghi\n")]);
+    let tree_a = create_tree(repo, &[(&path, "abc\ndef\nghi\n")]);
     let commit_a = tx
         .mut_repo()
         .new_commit(
             &settings,
             vec![repo.store().root_commit_id().clone()],
-            tree_a.id().clone(),
+            tree_a.id(),
         )
         .write()
         .unwrap();
-    let tree_b = testutils::create_tree(repo, &[(&path, "Abc\ndef\nghi\n")]);
+    let tree_b = create_tree(repo, &[(&path, "Abc\ndef\nghi\n")]);
     let commit_b = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_a.id().clone()], tree_b.id().clone())
+        .new_commit(&settings, vec![commit_a.id().clone()], tree_b.id())
         .write()
         .unwrap();
-    let tree_c = testutils::create_tree(repo, &[(&path, "Abc\ndef\nGhi\n")]);
+    let tree_c = create_tree(repo, &[(&path, "Abc\ndef\nGhi\n")]);
     let commit_c = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_b.id().clone()], tree_c.id().clone())
+        .new_commit(&settings, vec![commit_b.id().clone()], tree_c.id())
         .write()
         .unwrap();
-    let tree_d = testutils::create_tree(repo, &[(&path, "abC\ndef\nghi\n")]);
+    let tree_d = create_tree(repo, &[(&path, "abC\ndef\nghi\n")]);
     let commit_d = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_a.id().clone()], tree_d.id().clone())
+        .new_commit(&settings, vec![commit_a.id().clone()], tree_d.id())
         .write()
         .unwrap();
 
@@ -606,21 +604,17 @@ fn test_simplify_conflict_after_resolving_parent(use_git: bool) {
         rebase_commit(&settings, tx.mut_repo(), &commit_c, &[commit_b2.clone()]).unwrap();
 
     // Test the setup: Both B and C should have conflicts.
-    assert_matches!(
-        commit_b2.tree().path_value(&path),
-        Some(TreeValue::Conflict(_))
-    );
-    assert_matches!(
-        commit_c2.tree().path_value(&path),
-        Some(TreeValue::Conflict(_))
-    );
+    let tree_b2 = commit_b2.tree().unwrap();
+    let tree_c2 = commit_b2.tree().unwrap();
+    assert!(!tree_b2.path_value(&path).is_resolved());
+    assert!(!tree_c2.path_value(&path).is_resolved());
 
     // Create the resolved B and rebase C on top.
-    let tree_b3 = testutils::create_tree(repo, &[(&path, "AbC\ndef\nghi\n")]);
+    let tree_b3 = create_tree(repo, &[(&path, "AbC\ndef\nghi\n")]);
     let commit_b3 = tx
         .mut_repo()
         .rewrite_commit(&settings, &commit_b2)
-        .set_tree(tree_b3.id().clone())
+        .set_tree_id(tree_b3.id())
         .write()
         .unwrap();
     let commit_c3 = rebase_commit(&settings, tx.mut_repo(), &commit_c2, &[commit_b3]).unwrap();
@@ -628,12 +622,13 @@ fn test_simplify_conflict_after_resolving_parent(use_git: bool) {
     let repo = tx.commit();
 
     // The conflict should now be resolved.
-    let resolved_value = commit_c3.tree().path_value(&path);
-    match resolved_value {
-        Some(TreeValue::File {
+    let tree_c2 = commit_c3.tree().unwrap();
+    let resolved_value = tree_c2.path_value(&path);
+    match resolved_value.into_resolved() {
+        Ok(Some(TreeValue::File {
             id,
             executable: false,
-        }) => {
+        })) => {
             assert_eq!(
                 testutils::read_file(repo.store(), &path, &id),
                 b"AbC\ndef\nGhi\n"
