@@ -14,7 +14,7 @@
 
 use std::path::Path;
 
-use jj_lib::backend::CommitId;
+use jj_lib::backend::{CommitId, ObjectId};
 use jj_lib::repo::Repo;
 use test_case::test_case;
 use testutils::{create_random_commit, write_random_commit, TestRepo};
