@@ -20,7 +20,8 @@ use jj_lib::rewrite::DescendantRebaser;
 use maplit::{hashmap, hashset};
 use test_case::test_case;
 use testutils::{
-    assert_rebased, create_random_commit, write_random_commit, CommitGraphBuilder, TestRepo,
+    assert_rebased, create_random_commit, create_tree, write_random_commit, CommitGraphBuilder,
+    TestRepo,
 };
 
 #[test_case(false ; "local backend")]
@@ -856,35 +857,35 @@ fn test_rebase_descendants_contents(use_git: bool) {
     // A
     let mut tx = repo.start_transaction(&settings, "test");
     let path1 = RepoPath::from_internal_string("file1");
-    let tree1 = testutils::create_tree(repo, &[(&path1, "content")]);
+    let tree1 = create_tree(repo, &[(&path1, "content")]);
     let commit_a = tx
         .mut_repo()
         .new_commit(
             &settings,
             vec![repo.store().root_commit_id().clone()],
-            tree1.id().clone(),
+            tree1.id(),
         )
         .write()
         .unwrap();
     let path2 = RepoPath::from_internal_string("file2");
-    let tree2 = testutils::create_tree(repo, &[(&path2, "content")]);
+    let tree2 = create_tree(repo, &[(&path2, "content")]);
     let commit_b = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_a.id().clone()], tree2.id().clone())
+        .new_commit(&settings, vec![commit_a.id().clone()], tree2.id())
         .write()
         .unwrap();
     let path3 = RepoPath::from_internal_string("file3");
-    let tree3 = testutils::create_tree(repo, &[(&path3, "content")]);
+    let tree3 = create_tree(repo, &[(&path3, "content")]);
     let commit_c = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_b.id().clone()], tree3.id().clone())
+        .new_commit(&settings, vec![commit_b.id().clone()], tree3.id())
         .write()
         .unwrap();
     let path4 = RepoPath::from_internal_string("file4");
-    let tree4 = testutils::create_tree(repo, &[(&path4, "content")]);
+    let tree4 = create_tree(repo, &[(&path4, "content")]);
     let commit_d = tx
         .mut_repo()
-        .new_commit(&settings, vec![commit_a.id().clone()], tree4.id().clone())
+        .new_commit(&settings, vec![commit_a.id().clone()], tree4.id())
         .write()
         .unwrap();
 
@@ -904,18 +905,13 @@ fn test_rebase_descendants_contents(use_git: bool) {
         .get_commit(rebased.get(commit_c.id()).unwrap())
         .unwrap();
 
-    assert_eq!(
-        new_commit_c.tree().path_value(&path3),
-        commit_c.tree().path_value(&path3)
-    );
-    assert_eq!(
-        new_commit_c.tree().path_value(&path4),
-        commit_d.tree().path_value(&path4)
-    );
-    assert_ne!(
-        new_commit_c.tree().path_value(&path2),
-        commit_b.tree().path_value(&path2)
-    );
+    let tree_b = commit_b.tree().unwrap();
+    let tree_c = commit_c.tree().unwrap();
+    let tree_d = commit_d.tree().unwrap();
+    let new_tree_c = new_commit_c.tree().unwrap();
+    assert_eq!(new_tree_c.path_value(&path3), tree_c.path_value(&path3));
+    assert_eq!(new_tree_c.path_value(&path4), tree_d.path_value(&path4));
+    assert_ne!(new_tree_c.path_value(&path2), tree_b.path_value(&path2));
 }
 
 #[test]
@@ -1141,14 +1137,14 @@ fn test_rebase_descendants_update_branches_after_divergent_rewrite() {
         .unwrap();
     tx.mut_repo().rebase_descendants(&settings).unwrap();
 
-    let target = tx.mut_repo().get_local_branch("main").unwrap();
-    assert!(target.is_conflict());
+    let target = tx.mut_repo().get_local_branch("main");
+    assert!(target.has_conflict());
     assert_eq!(
-        target.removes().iter().counts(),
+        target.removed_ids().counts(),
         hashmap! { commit_b.id() => 2 },
     );
     assert_eq!(
-        target.adds().iter().counts(),
+        target.added_ids().counts(),
         hashmap! {
             commit_b2.id() => 1,
             commit_b3.id() => 1,
@@ -1216,14 +1212,14 @@ fn test_rebase_descendants_rewrite_updates_branch_conflict() {
         .unwrap();
     tx.mut_repo().rebase_descendants(&settings).unwrap();
 
-    let target = tx.mut_repo().get_local_branch("main").unwrap();
-    assert!(target.is_conflict());
+    let target = tx.mut_repo().get_local_branch("main");
+    assert!(target.has_conflict());
     assert_eq!(
-        target.removes().iter().counts(),
+        target.removed_ids().counts(),
         hashmap! { commit_a.id() => 1, commit_b.id() => 1 },
     );
     assert_eq!(
-        target.adds().iter().counts(),
+        target.added_ids().counts(),
         hashmap! {
             commit_c.id() => 1,
             commit_b2.id() => 1,
@@ -1313,7 +1309,7 @@ fn test_rebase_descendants_branch_delete_modify_abandon() {
     let mut tx = repo.start_transaction(&settings, "test");
     tx.mut_repo().record_abandoned_commit(commit_b.id().clone());
     tx.mut_repo().rebase_descendants(&settings).unwrap();
-    assert_eq!(tx.mut_repo().get_local_branch("main"), None);
+    assert_eq!(tx.mut_repo().get_local_branch("main"), RefTarget::absent());
 }
 
 #[test_case(false ; "local backend")]
