@@ -12,6 +12,10 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+// TODO: Remove when MSRV passes 1.72
+// https://github.com/frondeus/test-case/issues/126#issuecomment-1635916592
+#![allow(clippy::items_after_test_module)]
+
 use std::fs::OpenOptions;
 use std::io::Write;
 #[cfg(unix)]
@@ -21,19 +25,17 @@ use std::os::unix::net::UnixListener;
 use std::sync::Arc;
 
 use itertools::Itertools;
-use jj_lib::backend::{TreeId, TreeValue};
-use jj_lib::conflicts::Conflict;
+use jj_lib::backend::{MergedTreeId, ObjectId, TreeId, TreeValue};
 use jj_lib::fsmonitor::FsmonitorKind;
-#[cfg(unix)]
-use jj_lib::op_store::OperationId;
-use jj_lib::op_store::WorkspaceId;
+use jj_lib::merge::Merge;
+use jj_lib::merged_tree::{MergedTree, MergedTreeBuilder};
+use jj_lib::op_store::{OperationId, WorkspaceId};
 use jj_lib::repo::{ReadonlyRepo, Repo};
 use jj_lib::repo_path::{RepoPath, RepoPathComponent, RepoPathJoin};
 use jj_lib::settings::UserSettings;
-use jj_lib::tree_builder::TreeBuilder;
-use jj_lib::working_copy::{LockedWorkingCopy, SnapshotOptions, WorkingCopy};
+use jj_lib::working_copy::{LockedWorkingCopy, SnapshotError, SnapshotOptions, WorkingCopy};
 use test_case::test_case;
-use testutils::{write_random_commit, TestWorkspace};
+use testutils::{create_tree, write_random_commit, TestWorkspace};
 
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
@@ -41,22 +43,18 @@ fn test_root(use_git: bool) {
     // Test that the working copy is clean and empty after init.
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, use_git);
-    let repo = &test_workspace.repo;
 
-    let wc = test_workspace.workspace.working_copy_mut();
+    let wc = test_workspace.workspace.working_copy();
     assert_eq!(wc.sparse_patterns().unwrap(), vec![RepoPath::root()]);
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.discard();
+    let new_tree = test_workspace.snapshot().unwrap();
+    let repo = &test_workspace.repo;
     let wc_commit_id = repo
         .view()
         .get_wc_commit_id(&WorkspaceId::default())
         .unwrap();
     let wc_commit = repo.store().get_commit(wc_commit_id).unwrap();
-    assert_eq!(&new_tree_id, wc_commit.tree_id());
-    assert_eq!(&new_tree_id, repo.store().empty_tree_id());
+    assert_eq!(new_tree.id(), *wc_commit.tree_id());
+    assert_eq!(new_tree.id(), repo.store().empty_merged_tree_id());
 }
 
 #[test_case(false ; "local backend")]
@@ -89,41 +87,39 @@ fn test_checkout_file_transitions(use_git: bool) {
     fn write_path(
         settings: &UserSettings,
         repo: &Arc<ReadonlyRepo>,
-        tree_builder: &mut TreeBuilder,
+        tree_builder: &mut MergedTreeBuilder,
         kind: Kind,
         path: &RepoPath,
     ) {
         let store = repo.store();
         let value = match kind {
-            Kind::Missing => {
-                return;
-            }
+            Kind::Missing => Merge::absent(),
             Kind::Normal => {
                 let id = testutils::write_file(store, path, "normal file contents");
-                TreeValue::File {
+                Merge::normal(TreeValue::File {
                     id,
                     executable: false,
-                }
+                })
             }
             Kind::Executable => {
                 let id = testutils::write_file(store, path, "executable file contents");
-                TreeValue::File {
+                Merge::normal(TreeValue::File {
                     id,
                     executable: true,
-                }
+                })
             }
             Kind::ExecutableNormalContent => {
                 let id = testutils::write_file(store, path, "normal file contents");
-                TreeValue::File {
+                Merge::normal(TreeValue::File {
                     id,
                     executable: true,
-                }
+                })
             }
             Kind::Conflict => {
                 let base_file_id = testutils::write_file(store, path, "base file contents");
                 let left_file_id = testutils::write_file(store, path, "left file contents");
                 let right_file_id = testutils::write_file(store, path, "right file contents");
-                let conflict = Conflict::new(
+                Merge::new(
                     vec![Some(TreeValue::File {
                         id: base_file_id,
                         executable: false,
@@ -138,35 +134,30 @@ fn test_checkout_file_transitions(use_git: bool) {
                             executable: false,
                         }),
                     ],
-                );
-                let conflict_id = store.write_conflict(path, &conflict).unwrap();
-                TreeValue::Conflict(conflict_id)
+                )
             }
             Kind::Symlink => {
                 let id = store.write_symlink(path, "target").unwrap();
-                TreeValue::Symlink(id)
+                Merge::normal(TreeValue::Symlink(id))
             }
             Kind::Tree => {
-                let mut sub_tree_builder = store.tree_builder(store.empty_tree_id().clone());
                 let file_path = path.join(&RepoPathComponent::from("file"));
-                write_path(
-                    settings,
-                    repo,
-                    &mut sub_tree_builder,
-                    Kind::Normal,
-                    &file_path,
-                );
-                let id = sub_tree_builder.write_tree();
-                TreeValue::Tree(id)
+                let id = testutils::write_file(store, &file_path, "normal file contents");
+                let value = TreeValue::File {
+                    id,
+                    executable: false,
+                };
+                tree_builder.set_or_remove(file_path, Merge::normal(value));
+                return;
             }
             Kind::GitSubmodule => {
                 let mut tx = repo.start_transaction(settings, "test");
                 let id = write_random_commit(tx.mut_repo(), settings).id().clone();
                 tx.commit();
-                TreeValue::GitSubmodule(id)
+                Merge::normal(TreeValue::GitSubmodule(id))
             }
         };
-        tree_builder.set(path.clone(), value);
+        tree_builder.set_or_remove(path.clone(), value);
     }
 
     let mut kinds = vec![
@@ -182,8 +173,8 @@ fn test_checkout_file_transitions(use_git: bool) {
     if use_git {
         kinds.push(Kind::GitSubmodule);
     }
-    let mut left_tree_builder = store.tree_builder(store.empty_tree_id().clone());
-    let mut right_tree_builder = store.tree_builder(store.empty_tree_id().clone());
+    let mut left_tree_builder = MergedTreeBuilder::new(store.empty_merged_tree_id());
+    let mut right_tree_builder = MergedTreeBuilder::new(store.empty_merged_tree_id());
     let mut files = vec![];
     for left_kind in &kinds {
         for right_kind in &kinds {
@@ -193,10 +184,10 @@ fn test_checkout_file_transitions(use_git: bool) {
             files.push((*left_kind, *right_kind, path));
         }
     }
-    let left_tree_id = left_tree_builder.write_tree();
-    let right_tree_id = right_tree_builder.write_tree();
-    let left_tree = store.get_tree(&RepoPath::root(), &left_tree_id).unwrap();
-    let right_tree = store.get_tree(&RepoPath::root(), &right_tree_id).unwrap();
+    let left_tree_id = left_tree_builder.write_tree(&store).unwrap();
+    let right_tree_id = right_tree_builder.write_tree(&store).unwrap();
+    let left_tree = store.get_root_tree(&left_tree_id).unwrap();
+    let right_tree = store.get_root_tree(&right_tree_id).unwrap();
 
     let wc = test_workspace.workspace.working_copy_mut();
     wc.check_out(repo.op_id().clone(), None, &left_tree)
@@ -205,12 +196,8 @@ fn test_checkout_file_transitions(use_git: bool) {
         .unwrap();
 
     // Check that the working copy is clean.
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.discard();
-    assert_eq!(new_tree_id, right_tree_id);
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(new_tree.id(), right_tree_id);
 
     for (_left_kind, right_kind, path) in &files {
         let wc_path = workspace_root.join(path.to_internal_file_string());
@@ -273,6 +260,25 @@ fn test_checkout_file_transitions(use_git: bool) {
     }
 }
 
+// Test case for issue #2165
+#[test]
+fn test_conflict_subdirectory() {
+    let settings = testutils::user_settings();
+    let mut test_workspace = TestWorkspace::init(&settings, true);
+    let repo = &test_workspace.repo;
+
+    let path = RepoPath::from_internal_string("sub/file");
+    let empty_tree = create_tree(repo, &[]);
+    let tree1 = create_tree(repo, &[(&path, "0")]);
+    let tree2 = create_tree(repo, &[(&path, "1")]);
+    let merged_tree = tree1.merge(&empty_tree, &tree2).unwrap();
+    let repo = &test_workspace.repo;
+    let wc = test_workspace.workspace.working_copy_mut();
+    wc.check_out(repo.op_id().clone(), None, &tree1).unwrap();
+    wc.check_out(repo.op_id().clone(), None, &merged_tree)
+        .unwrap();
+}
+
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
 fn test_tree_builder_file_directory_transition(use_git: bool) {
@@ -285,7 +291,8 @@ fn test_tree_builder_file_directory_transition(use_git: bool) {
     let mut check_out_tree = |tree_id: &TreeId| {
         let tree = repo.store().get_tree(&RepoPath::root(), tree_id).unwrap();
         let wc = workspace.working_copy_mut();
-        wc.check_out(repo.op_id().clone(), None, &tree).unwrap();
+        wc.check_out(repo.op_id().clone(), None, &MergedTree::legacy(tree))
+            .unwrap();
     };
 
     let parent_path = RepoPath::from_internal_string("foo/bar");
@@ -323,13 +330,14 @@ fn test_reset() {
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, false);
     let repo = &test_workspace.repo;
+    let op_id = repo.op_id().clone();
     let workspace_root = test_workspace.workspace.workspace_root().clone();
 
     let ignored_path = RepoPath::from_internal_string("ignored");
     let gitignore_path = RepoPath::from_internal_string(".gitignore");
 
-    let tree_without_file = testutils::create_tree(repo, &[(&gitignore_path, "ignored\n")]);
-    let tree_with_file = testutils::create_tree(
+    let tree_without_file = create_tree(repo, &[(&gitignore_path, "ignored\n")]);
+    let tree_with_file = create_tree(
         repo,
         &[(&gitignore_path, "ignored\n"), (&ignored_path, "code")],
     );
@@ -347,44 +355,22 @@ fn test_reset() {
     // commit the working copy (because it's ignored).
     let mut locked_wc = wc.start_mutation().unwrap();
     locked_wc.reset(&tree_without_file).unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
+    locked_wc.finish(op_id.clone()).unwrap();
     assert!(ignored_path.to_fs_path(&workspace_root).is_file());
     assert!(!wc.file_states().unwrap().contains_key(&ignored_path));
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    assert_eq!(new_tree_id, *tree_without_file.id());
-    locked_wc.discard();
-
-    // After we reset to the commit without the file, it should still exist on disk,
-    // but it should not be in the tree state, and it should not get added when we
-    // commit the working copy (because it's ignored).
-    let mut locked_wc = wc.start_mutation().unwrap();
-    locked_wc.reset(&tree_without_file).unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
-    assert!(ignored_path.to_fs_path(&workspace_root).is_file());
-    assert!(!wc.file_states().unwrap().contains_key(&ignored_path));
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    assert_eq!(new_tree_id, *tree_without_file.id());
-    locked_wc.discard();
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(new_tree.id(), tree_without_file.id());
 
     // Now test the opposite direction: resetting to a commit where the file is
     // tracked. The file should become tracked (even though it's ignored).
+    let wc = test_workspace.workspace.working_copy_mut();
     let mut locked_wc = wc.start_mutation().unwrap();
     locked_wc.reset(&tree_with_file).unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
+    locked_wc.finish(op_id.clone()).unwrap();
     assert!(ignored_path.to_fs_path(&workspace_root).is_file());
     assert!(wc.file_states().unwrap().contains_key(&ignored_path));
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    assert_eq!(new_tree_id, *tree_with_file.id());
-    locked_wc.discard();
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(new_tree.id(), tree_with_file.id());
 }
 
 #[test]
@@ -401,8 +387,8 @@ fn test_checkout_discard() {
     let file2_path = RepoPath::from_internal_string("file2");
 
     let store = repo.store();
-    let tree1 = testutils::create_tree(&repo, &[(&file1_path, "contents")]);
-    let tree2 = testutils::create_tree(&repo, &[(&file2_path, "contents")]);
+    let tree1 = create_tree(&repo, &[(&file1_path, "contents")]);
+    let tree2 = create_tree(&repo, &[(&file2_path, "contents")]);
 
     let wc = test_workspace.workspace.working_copy_mut();
     let state_path = wc.state_path().to_path_buf();
@@ -421,7 +407,7 @@ fn test_checkout_discard() {
     let reloaded_wc = WorkingCopy::load(store.clone(), workspace_root.clone(), state_path.clone());
     assert!(reloaded_wc.file_states().unwrap().contains_key(&file1_path));
     assert!(!reloaded_wc.file_states().unwrap().contains_key(&file2_path));
-    locked_wc.discard();
+    drop(locked_wc);
 
     // The change should remain in the working copy, but not in memory and not saved
     assert!(wc.file_states().unwrap().contains_key(&file1_path));
@@ -444,7 +430,7 @@ fn test_snapshot_racy_timestamps(use_git: bool) {
     let workspace_root = test_workspace.workspace.workspace_root().clone();
 
     let file_path = workspace_root.join("file");
-    let mut previous_tree_id = repo.store().empty_tree_id().clone();
+    let mut previous_tree_id = repo.store().empty_merged_tree_id();
     let wc = test_workspace.workspace.working_copy_mut();
     for i in 0..100 {
         {
@@ -461,7 +447,6 @@ fn test_snapshot_racy_timestamps(use_git: bool) {
         let new_tree_id = locked_wc
             .snapshot(SnapshotOptions::empty_for_test())
             .unwrap();
-        locked_wc.discard();
         assert_ne!(new_tree_id, previous_tree_id);
         previous_tree_id = new_tree_id;
     }
@@ -496,7 +481,7 @@ fn test_snapshot_special_file() {
         .snapshot(SnapshotOptions::empty_for_test())
         .unwrap();
     locked_wc.finish(OperationId::from_hex("abc123")).unwrap();
-    let tree = store.get_tree(&RepoPath::root(), &tree_id).unwrap();
+    let tree = store.get_root_tree(&tree_id).unwrap();
     // Only the regular files should be in the tree
     assert_eq!(
         tree.entries().map(|(path, _value)| path).collect_vec(),
@@ -510,17 +495,13 @@ fn test_snapshot_special_file() {
     // Replace a regular file by a socket and snapshot the working copy again
     std::fs::remove_file(&file1_disk_path).unwrap();
     UnixListener::bind(&file1_disk_path).unwrap();
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.finish(OperationId::from_hex("abc123")).unwrap();
-    let tree = store.get_tree(&RepoPath::root(), &tree_id).unwrap();
+    let tree = test_workspace.snapshot().unwrap();
     // Only the regular file should be in the tree
     assert_eq!(
         tree.entries().map(|(path, _value)| path).collect_vec(),
         vec![file2_path.clone()]
     );
+    let wc = test_workspace.workspace.working_copy();
     assert_eq!(
         wc.file_states().unwrap().keys().cloned().collect_vec(),
         vec![file2_path]
@@ -534,7 +515,6 @@ fn test_gitignores(use_git: bool) {
 
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, use_git);
-    let repo = &test_workspace.repo;
     let workspace_root = test_workspace.workspace.workspace_root().clone();
 
     let gitignore_path = RepoPath::from_internal_string(".gitignore");
@@ -551,16 +531,7 @@ fn test_gitignores(use_git: bool) {
     std::fs::create_dir(workspace_root.join("dir")).unwrap();
     testutils::write_working_copy_file(&workspace_root, &subdir_modified_path, "1");
 
-    let wc = test_workspace.workspace.working_copy_mut();
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id1 = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
-    let tree1 = repo
-        .store()
-        .get_tree(&RepoPath::root(), &new_tree_id1)
-        .unwrap();
+    let tree1 = test_workspace.snapshot().unwrap();
     let files1 = tree1.entries().map(|(name, _value)| name).collect_vec();
     assert_eq!(
         files1,
@@ -584,15 +555,7 @@ fn test_gitignores(use_git: bool) {
     testutils::write_working_copy_file(&workspace_root, &subdir_modified_path, "2");
     testutils::write_working_copy_file(&workspace_root, &subdir_ignored_path, "2");
 
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id2 = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.discard();
-    let tree2 = repo
-        .store()
-        .get_tree(&RepoPath::root(), &new_tree_id2)
-        .unwrap();
+    let tree2 = test_workspace.snapshot().unwrap();
     let files2 = tree2.entries().map(|(name, _value)| name).collect_vec();
     assert_eq!(
         files2,
@@ -605,6 +568,54 @@ fn test_gitignores(use_git: bool) {
     );
 }
 
+#[test_case(false ; "local backend")]
+#[test_case(true ; "git backend")]
+fn test_gitignores_in_ignored_dir(use_git: bool) {
+    // Tests that .gitignore files in an ignored directory are ignored, i.e. that
+    // they cannot override the ignores from the parent
+
+    let settings = testutils::user_settings();
+    let mut test_workspace = TestWorkspace::init(&settings, use_git);
+    let op_id = test_workspace.repo.op_id().clone();
+    let workspace_root = test_workspace.workspace.workspace_root().clone();
+
+    let gitignore_path = RepoPath::from_internal_string(".gitignore");
+    let nested_gitignore_path = RepoPath::from_internal_string("ignored/.gitignore");
+    let ignored_path = RepoPath::from_internal_string("ignored/file");
+
+    let tree1 = create_tree(&test_workspace.repo, &[(&gitignore_path, "ignored\n")]);
+    let wc = test_workspace.workspace.working_copy_mut();
+    wc.check_out(op_id.clone(), None, &tree1).unwrap();
+
+    testutils::write_working_copy_file(&workspace_root, &nested_gitignore_path, "!file\n");
+    testutils::write_working_copy_file(&workspace_root, &ignored_path, "contents");
+
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(
+        new_tree.entries().collect_vec(),
+        tree1.entries().collect_vec()
+    );
+
+    // The nested .gitignore is ignored even if it's tracked
+    let tree2 = create_tree(
+        &test_workspace.repo,
+        &[
+            (&gitignore_path, "ignored\n"),
+            (&nested_gitignore_path, "!file\n"),
+        ],
+    );
+    let wc = test_workspace.workspace.working_copy_mut();
+    let mut locked_wc = wc.start_mutation().unwrap();
+    locked_wc.reset(&tree2).unwrap();
+    locked_wc.finish(OperationId::from_hex("abc123")).unwrap();
+
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(
+        new_tree.entries().collect_vec(),
+        tree2.entries().collect_vec()
+    );
+}
+
 #[test_case(false ; "local backend")]
 #[test_case(true ; "git backend")]
 fn test_gitignores_checkout_never_overwrites_ignored(use_git: bool) {
@@ -623,12 +634,7 @@ fn test_gitignores_checkout_never_overwrites_ignored(use_git: bool) {
     testutils::write_working_copy_file(&workspace_root, &modified_path, "garbage");
 
     // Create a tree that adds the same file but with different contents
-    let mut tree_builder = repo
-        .store()
-        .tree_builder(repo.store().empty_tree_id().clone());
-    testutils::write_normal_file(&mut tree_builder, &modified_path, "contents");
-    let tree_id = tree_builder.write_tree();
-    let tree = repo.store().get_tree(&RepoPath::root(), &tree_id).unwrap();
+    let tree = create_tree(repo, &[(&modified_path, "contents")]);
 
     // Now check out the tree that adds the file "modified" with contents
     // "contents". The exiting contents ("garbage") shouldn't be replaced in the
@@ -650,41 +656,45 @@ fn test_gitignores_ignored_directory_already_tracked(use_git: bool) {
 
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, use_git);
-    let repo = &test_workspace.repo;
+    let workspace_root = test_workspace.workspace.workspace_root().clone();
+    let repo = test_workspace.repo.clone();
 
-    // Add a .gitignore file saying to ignore the directory "ignored/"
     let gitignore_path = RepoPath::from_internal_string(".gitignore");
-    testutils::write_working_copy_file(
-        test_workspace.workspace.workspace_root(),
-        &gitignore_path,
-        "/ignored/\n",
+    let unchanged_path = RepoPath::from_internal_string("ignored/unchanged");
+    let modified_path = RepoPath::from_internal_string("ignored/modified");
+    let deleted_path = RepoPath::from_internal_string("ignored/deleted");
+    let tree = create_tree(
+        &repo,
+        &[
+            (&gitignore_path, "/ignored/\n"),
+            (&unchanged_path, "contents"),
+            (&modified_path, "contents"),
+            (&deleted_path, "contents"),
+        ],
     );
-    let file_path = RepoPath::from_internal_string("ignored/file");
 
-    // Create a tree that adds a file in the ignored directory
-    let mut tree_builder = repo
-        .store()
-        .tree_builder(repo.store().empty_tree_id().clone());
-    testutils::write_normal_file(&mut tree_builder, &file_path, "contents");
-    let tree_id = tree_builder.write_tree();
-    let tree = repo.store().get_tree(&RepoPath::root(), &tree_id).unwrap();
-
-    // Check out the tree with the file in ignored/
+    // Check out the tree with the files in `ignored/`
     let wc = test_workspace.workspace.working_copy_mut();
     wc.check_out(repo.op_id().clone(), None, &tree).unwrap();
 
-    // Check that the file is still in the tree created by snapshotting the working
-    // copy (that it didn't get removed because the directory is ignored)
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.discard();
-    let new_tree = repo
-        .store()
-        .get_tree(&RepoPath::root(), &new_tree_id)
-        .unwrap();
-    assert!(new_tree.path_value(&file_path).is_some());
+    // Make some changes inside the ignored directory and check that they are
+    // detected when we snapshot. The files that are still there should not be
+    // deleted from the resulting tree.
+    std::fs::write(modified_path.to_fs_path(&workspace_root), "modified").unwrap();
+    std::fs::remove_file(deleted_path.to_fs_path(&workspace_root)).unwrap();
+    let new_tree = test_workspace.snapshot().unwrap();
+    let expected_tree = create_tree(
+        &repo,
+        &[
+            (&gitignore_path, "/ignored/\n"),
+            (&unchanged_path, "contents"),
+            (&modified_path, "modified"),
+        ],
+    );
+    assert_eq!(
+        new_tree.entries().collect_vec(),
+        expected_tree.entries().collect_vec()
+    );
 }
 
 #[test_case(false ; "local backend")]
@@ -695,7 +705,7 @@ fn test_dotgit_ignored(use_git: bool) {
 
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, use_git);
-    let repo = &test_workspace.repo;
+    let store = test_workspace.repo.store().clone();
     let workspace_root = test_workspace.workspace.workspace_root().clone();
 
     // Test with a .git/ directory (with a file in, since we don't write empty
@@ -707,16 +717,9 @@ fn test_dotgit_ignored(use_git: bool) {
         &RepoPath::from_internal_string(".git/file"),
         "contents",
     );
-    let mut locked_wc = test_workspace
-        .workspace
-        .working_copy_mut()
-        .start_mutation()
-        .unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    assert_eq!(new_tree_id, *repo.store().empty_tree_id());
-    locked_wc.discard();
+    let new_tree = test_workspace.snapshot().unwrap();
+    let empty_tree_id = store.empty_merged_tree_id();
+    assert_eq!(new_tree.id(), empty_tree_id);
     std::fs::remove_dir_all(&dotgit_path).unwrap();
 
     // Test with a .git file
@@ -725,16 +728,8 @@ fn test_dotgit_ignored(use_git: bool) {
         &RepoPath::from_internal_string(".git"),
         "contents",
     );
-    let mut locked_wc = test_workspace
-        .workspace
-        .working_copy_mut()
-        .start_mutation()
-        .unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    assert_eq!(new_tree_id, *repo.store().empty_tree_id());
-    locked_wc.discard();
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(new_tree.id(), empty_tree_id);
 }
 
 #[test]
@@ -770,8 +765,8 @@ fn test_gitsubmodule() {
         TreeValue::GitSubmodule(submodule_id),
     );
 
-    let tree_id = tree_builder.write_tree();
-    let tree = store.get_tree(&RepoPath::root(), &tree_id).unwrap();
+    let tree_id = MergedTreeId::Legacy(tree_builder.write_tree());
+    let tree = store.get_root_tree(&tree_id).unwrap();
     let wc = test_workspace.workspace.working_copy_mut();
     wc.check_out(repo.op_id().clone(), None, &tree).unwrap();
 
@@ -785,12 +780,8 @@ fn test_gitsubmodule() {
 
     // Check that the files present in the submodule are not tracked
     // when we snapshot
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let new_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.discard();
-    assert_eq!(new_tree_id, tree_id);
+    let new_tree = test_workspace.snapshot().unwrap();
+    assert_eq!(new_tree.id(), tree_id);
 
     // Check that the files in the submodule are not deleted
     let file_in_submodule_path = added_submodule_path.to_fs_path(&workspace_root);
@@ -812,16 +803,8 @@ fn test_existing_directory_symlink(use_git: bool) {
     // Creates a symlink in working directory, and a tree that will add a file under
     // the symlinked directory.
     std::os::unix::fs::symlink("..", workspace_root.join("parent")).unwrap();
-    let mut tree_builder = repo
-        .store()
-        .tree_builder(repo.store().empty_tree_id().clone());
-    testutils::write_normal_file(
-        &mut tree_builder,
-        &RepoPath::from_internal_string("parent/escaped"),
-        "contents",
-    );
-    let tree_id = tree_builder.write_tree();
-    let tree = repo.store().get_tree(&RepoPath::root(), &tree_id).unwrap();
+    let file_path = RepoPath::from_internal_string("parent/escaped");
+    let tree = create_tree(repo, &[(&file_path, "contents")]);
 
     // Checkout should fail because "parent" already exists and is a symlink.
     let wc = test_workspace.workspace.working_copy_mut();
@@ -871,8 +854,7 @@ fn test_fsmonitor() {
     {
         let mut locked_wc = wc.start_mutation().unwrap();
         let tree_id = snapshot(&mut locked_wc, &[]);
-        assert_eq!(tree_id, *repo.store().empty_tree_id());
-        locked_wc.discard();
+        assert_eq!(tree_id, repo.store().empty_merged_tree_id());
     }
 
     {
@@ -882,7 +864,6 @@ fn test_fsmonitor() {
         tree 205f6b799e7d5c2524468ca006a0131aa57ecce7
           file "foo" (257cc5642cb1a054f08cc83f2d943e56fd3ebe99): "foo\n"
         "###);
-        locked_wc.discard();
     }
 
     {
@@ -911,7 +892,6 @@ fn test_fsmonitor() {
           file "foo" (9d053d7c8a18a286dce9b99a59bb058be173b463): "updated foo\n"
           file "path/to/nested" (79c53955ef856f16f2107446bc721c8879a1bd2e): "nested\n"
         "###);
-        locked_wc.discard();
     }
 
     {
@@ -926,3 +906,37 @@ fn test_fsmonitor() {
         locked_wc.finish(repo.op_id().clone()).unwrap();
     }
 }
+
+#[test]
+fn test_snapshot_max_new_file_size() {
+    let settings = UserSettings::from_config(
+        testutils::base_config()
+            .add_source(config::File::from_str(
+                "snapshot.max-new-file-size = \"1KiB\"",
+                config::FileFormat::Toml,
+            ))
+            .build()
+            .unwrap(),
+    );
+    let mut test_workspace = TestWorkspace::init(&settings, true);
+    let workspace_root = test_workspace.workspace.workspace_root().clone();
+    let small_path = RepoPath::from_internal_string("small");
+    let large_path = RepoPath::from_internal_string("large");
+    std::fs::write(small_path.to_fs_path(&workspace_root), vec![0; 1024]).unwrap();
+    test_workspace
+        .snapshot()
+        .expect("files exactly matching the size limit should succeed");
+    std::fs::write(small_path.to_fs_path(&workspace_root), vec![0; 1024 + 1]).unwrap();
+    test_workspace
+        .snapshot()
+        .expect("existing files may grow beyond the size limit");
+    // A new file of 1KiB + 1 bytes should fail
+    std::fs::write(large_path.to_fs_path(&workspace_root), vec![0; 1024 + 1]).unwrap();
+    let err = test_workspace
+        .snapshot()
+        .expect_err("new files beyond the size limit should fail");
+    assert!(
+        matches!(err, SnapshotError::NewFileTooLarge { .. }),
+        "the failure should be attributed to new file size"
+    );
+}
