@@ -20,34 +20,23 @@ use jj_lib::repo::{Repo, StoreFactories};
 use jj_lib::repo_path::RepoPath;
 use jj_lib::working_copy::{CheckoutError, SnapshotOptions};
 use jj_lib::workspace::Workspace;
-use test_case::test_case;
-use testutils::TestWorkspace;
+use testutils::{create_tree, write_working_copy_file, TestWorkspace};
 
-#[test_case(false ; "local backend")]
-#[test_case(true ; "git backend")]
-fn test_concurrent_checkout(use_git: bool) {
+#[test]
+fn test_concurrent_checkout() {
     // Test that we error out if a concurrent checkout is detected (i.e. if the
     // working-copy commit changed on disk after we read it).
     let settings = testutils::user_settings();
-    let mut test_workspace1 = TestWorkspace::init(&settings, use_git);
+    let mut test_workspace1 = TestWorkspace::init(&settings, true);
     let repo1 = test_workspace1.repo.clone();
     let workspace1_root = test_workspace1.workspace.workspace_root().clone();
 
     let tree_id1 = testutils::create_random_tree(&repo1);
     let tree_id2 = testutils::create_random_tree(&repo1);
     let tree_id3 = testutils::create_random_tree(&repo1);
-    let tree1 = repo1
-        .store()
-        .get_tree(&RepoPath::root(), &tree_id1)
-        .unwrap();
-    let tree2 = repo1
-        .store()
-        .get_tree(&RepoPath::root(), &tree_id2)
-        .unwrap();
-    let tree3 = repo1
-        .store()
-        .get_tree(&RepoPath::root(), &tree_id3)
-        .unwrap();
+    let tree1 = repo1.store().get_root_tree(&tree_id1).unwrap();
+    let tree2 = repo1.store().get_root_tree(&tree_id2).unwrap();
+    let tree3 = repo1.store().get_root_tree(&tree_id3).unwrap();
 
     // Check out tree1
     let wc1 = test_workspace1.workspace.working_copy_mut();
@@ -73,18 +62,17 @@ fn test_concurrent_checkout(use_git: bool) {
     let workspace3 =
         Workspace::load(&settings, &workspace1_root, &StoreFactories::default()).unwrap();
     assert_eq!(
-        workspace3.working_copy().current_tree_id().unwrap(),
-        &tree_id2
+        *workspace3.working_copy().current_tree_id().unwrap(),
+        tree_id2
     );
 }
 
-#[test_case(false ; "local backend")]
-#[test_case(true ; "git backend")]
-fn test_checkout_parallel(use_git: bool) {
+#[test]
+fn test_checkout_parallel() {
     // Test that concurrent checkouts by different processes (simulated by using
     // different repo instances) is safe.
     let settings = testutils::user_settings();
-    let mut test_workspace = TestWorkspace::init(&settings, use_git);
+    let mut test_workspace = TestWorkspace::init(&settings, true);
     let repo = &test_workspace.repo;
     let workspace_root = test_workspace.workspace.workspace_root().clone();
 
@@ -92,13 +80,13 @@ fn test_checkout_parallel(use_git: bool) {
     let mut tree_ids = vec![];
     for i in 0..num_threads {
         let path = RepoPath::from_internal_string(format!("file{i}").as_str());
-        let tree = testutils::create_tree(repo, &[(&path, "contents")]);
-        tree_ids.push(tree.id().clone());
+        let tree = create_tree(repo, &[(&path, "contents")]);
+        tree_ids.push(tree.id());
     }
 
     // Create another tree just so we can test the update stats reliably from the
     // first update
-    let tree = testutils::create_tree(
+    let tree = create_tree(
         repo,
         &[(&RepoPath::from_internal_string("other file"), "contents")],
     );
@@ -122,7 +110,7 @@ fn test_checkout_parallel(use_git: bool) {
                 let tree = workspace
                     .repo_loader()
                     .store()
-                    .get_tree(&RepoPath::root(), &tree_id)
+                    .get_root_tree(&tree_id)
                     .unwrap();
                 // The operation ID is not correct, but that doesn't matter for this test
                 let stats = workspace
@@ -140,9 +128,40 @@ fn test_checkout_parallel(use_git: bool) {
                 let new_tree_id = locked_wc
                     .snapshot(SnapshotOptions::empty_for_test())
                     .unwrap();
-                locked_wc.discard();
                 assert!(tree_ids.contains(&new_tree_id));
             });
         }
     });
 }
+
+#[test]
+fn test_racy_checkout() {
+    let settings = testutils::user_settings();
+    let mut test_workspace = TestWorkspace::init(&settings, true);
+    let repo = &test_workspace.repo;
+    let op_id = repo.op_id().clone();
+    let workspace_root = test_workspace.workspace.workspace_root().clone();
+
+    let path = RepoPath::from_internal_string("file");
+    let tree = create_tree(repo, &[(&path, "1")]);
+
+    let mut num_matches = 0;
+    for _ in 0..100 {
+        let wc = test_workspace.workspace.working_copy_mut();
+        wc.check_out(op_id.clone(), None, &tree).unwrap();
+        assert_eq!(
+            std::fs::read(path.to_fs_path(&workspace_root)).unwrap(),
+            b"1".to_vec()
+        );
+        // A file written right after checkout (hopefully, from the test's perspective,
+        // within the file system timestamp granularity) is detected as changed.
+        write_working_copy_file(&workspace_root, &path, "x");
+        let modified_tree = test_workspace.snapshot().unwrap();
+        if modified_tree.id() == tree.id() {
+            num_matches += 1;
+        }
+        // Reset the state for the next round
+        write_working_copy_file(&workspace_root, &path, "1");
+    }
+    assert_eq!(num_matches, 0);
+}
