@@ -16,8 +16,8 @@ use itertools::Itertools;
 use jj_lib::matchers::EverythingMatcher;
 use jj_lib::repo::Repo;
 use jj_lib::repo_path::RepoPath;
-use jj_lib::working_copy::{CheckoutStats, SnapshotOptions, WorkingCopy};
-use testutils::TestWorkspace;
+use jj_lib::working_copy::{CheckoutStats, WorkingCopy};
+use testutils::{create_tree, TestWorkspace};
 
 #[test]
 fn test_sparse_checkout() {
@@ -36,7 +36,7 @@ fn test_sparse_checkout() {
     let dir2_path = RepoPath::from_internal_string("dir2");
     let dir2_file1_path = RepoPath::from_internal_string("dir2/file1");
 
-    let tree = testutils::create_tree(
+    let tree = create_tree(
         repo,
         &[
             (&root_file1_path, "contents"),
@@ -131,6 +131,7 @@ fn test_sparse_commit() {
     let settings = testutils::user_settings();
     let mut test_workspace = TestWorkspace::init(&settings, false);
     let repo = &test_workspace.repo;
+    let op_id = repo.op_id().clone();
     let working_copy_path = test_workspace.workspace.workspace_root().clone();
 
     let root_file1_path = RepoPath::from_internal_string("file1");
@@ -139,7 +140,7 @@ fn test_sparse_commit() {
     let dir2_path = RepoPath::from_internal_string("dir2");
     let dir2_file1_path = RepoPath::from_internal_string("dir2/file1");
 
-    let tree = testutils::create_tree(
+    let tree = create_tree(
         repo,
         &[
             (&root_file1_path, "contents"),
@@ -166,36 +167,21 @@ fn test_sparse_commit() {
 
     // Create a tree from the working copy. Only dir1/file1 should be updated in the
     // tree.
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let modified_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
-    let modified_tree = repo
-        .store()
-        .get_tree(&RepoPath::root(), &modified_tree_id)
-        .unwrap();
+    let modified_tree = test_workspace.snapshot().unwrap();
     let diff = tree.diff(&modified_tree, &EverythingMatcher).collect_vec();
     assert_eq!(diff.len(), 1);
     assert_eq!(diff[0].0, dir1_file1_path);
 
     // Set sparse patterns to also include dir2/
+    let wc = test_workspace.workspace.working_copy_mut();
     let mut locked_wc = wc.start_mutation().unwrap();
     let sparse_patterns = vec![dir1_path, dir2_path];
     locked_wc.set_sparse_patterns(sparse_patterns).unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
+    locked_wc.finish(op_id).unwrap();
 
     // Create a tree from the working copy. Only dir1/file1 and dir2/file1 should be
     // updated in the tree.
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let modified_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
-    let modified_tree = repo
-        .store()
-        .get_tree(&RepoPath::root(), &modified_tree_id)
-        .unwrap();
+    let modified_tree = test_workspace.snapshot().unwrap();
     let diff = tree.diff(&modified_tree, &EverythingMatcher).collect_vec();
     assert_eq!(diff.len(), 2);
     assert_eq!(diff[0].0, dir1_file1_path);
@@ -230,15 +216,7 @@ fn test_sparse_commit_gitignore() {
 
     // Create a tree from the working copy. Only dir1/file2 should be updated in the
     // tree because dir1/file1 is ignored.
-    let mut locked_wc = wc.start_mutation().unwrap();
-    let modified_tree_id = locked_wc
-        .snapshot(SnapshotOptions::empty_for_test())
-        .unwrap();
-    locked_wc.finish(repo.op_id().clone()).unwrap();
-    let modified_tree = repo
-        .store()
-        .get_tree(&RepoPath::root(), &modified_tree_id)
-        .unwrap();
+    let modified_tree = test_workspace.snapshot().unwrap();
     let entries = modified_tree.entries().collect_vec();
     assert_eq!(entries.len(), 1);
     assert_eq!(entries[0].0, dir1_file2_path);
