@@ -18,18 +18,21 @@ use std::path::{Path, PathBuf};
 use std::sync::{Arc, Once};
 
 use itertools::Itertools;
-use jj_lib::backend::{Backend, BackendInitError, FileId, ObjectId, TreeId, TreeValue};
+use jj_lib::backend::{Backend, BackendInitError, FileId, MergedTreeId, ObjectId, TreeValue};
 use jj_lib::commit::Commit;
 use jj_lib::commit_builder::CommitBuilder;
 use jj_lib::git_backend::GitBackend;
 use jj_lib::local_backend::LocalBackend;
+use jj_lib::merged_tree::MergedTree;
 use jj_lib::repo::{MutableRepo, ReadonlyRepo, Repo, RepoLoader, StoreFactories};
 use jj_lib::repo_path::RepoPath;
 use jj_lib::rewrite::RebasedDescendant;
 use jj_lib::settings::UserSettings;
 use jj_lib::store::Store;
+use jj_lib::transaction::Transaction;
 use jj_lib::tree::Tree;
 use jj_lib::tree_builder::TreeBuilder;
+use jj_lib::working_copy::{SnapshotError, SnapshotOptions};
 use jj_lib::workspace::Workspace;
 use tempfile::TempDir;
 
@@ -58,20 +61,21 @@ pub fn new_temp_dir() -> TempDir {
         .unwrap()
 }
 
+pub fn base_config() -> config::ConfigBuilder<config::builder::DefaultState> {
+    config::Config::builder().add_source(config::File::from_str(
+        r#"
+            user.name = "Test User"
+            user.email = "test.user@example.com"
+            operation.username = "test-username"
+            operation.hostname = "host.example.com"
+            debug.randomness-seed = "42"
+        "#,
+        config::FileFormat::Toml,
+    ))
+}
+
 pub fn user_settings() -> UserSettings {
-    let config = config::Config::builder()
-        .add_source(config::File::from_str(
-            r#"
-                user.name = "Test User"
-                user.email = "test.user@example.com"
-                operation.username = "test-username"
-                operation.hostname = "host.example.com"
-                debug.randomness-seed = "42"
-           "#,
-            config::FileFormat::Toml,
-        ))
-        .build()
-        .unwrap();
+    let config = base_config().build().unwrap();
     UserSettings::from_config(config)
 }
 
@@ -129,6 +133,7 @@ pub struct TestWorkspace {
     temp_dir: TempDir,
     pub workspace: Workspace,
     pub repo: Arc<ReadonlyRepo>,
+    settings: UserSettings,
 }
 
 impl TestWorkspace {
@@ -150,12 +155,27 @@ impl TestWorkspace {
             temp_dir,
             workspace,
             repo,
+            settings: settings.clone(),
         }
     }
 
     pub fn root_dir(&self) -> PathBuf {
         self.temp_dir.path().join("repo").join("..")
     }
+
+    /// Snapshots the working copy and returns the tree. Updates the working
+    /// copy state on disk, but does not update the working-copy commit (no
+    /// new operation).
+    pub fn snapshot(&mut self) -> Result<MergedTree, SnapshotError> {
+        let mut locked_wc = self.workspace.working_copy_mut().start_mutation().unwrap();
+        let tree_id = locked_wc.snapshot(SnapshotOptions {
+            max_new_file_size: self.settings.max_new_file_size().unwrap(),
+            ..SnapshotOptions::empty_for_test()
+        })?;
+        // arbitrary operation id
+        locked_wc.finish(self.repo.op_id().clone()).unwrap();
+        Ok(self.repo.store().get_root_tree(&tree_id).unwrap())
+    }
 }
 
 pub fn load_repo_at_head(settings: &UserSettings, repo_path: &Path) -> Arc<ReadonlyRepo> {
@@ -165,6 +185,23 @@ pub fn load_repo_at_head(settings: &UserSettings, repo_path: &Path) -> Arc<Reado
         .unwrap()
 }
 
+pub fn commit_transactions(settings: &UserSettings, txs: Vec<Transaction>) -> Arc<ReadonlyRepo> {
+    let repo_loader = txs[0].base_repo().loader();
+    let mut op_ids = vec![];
+    for tx in txs {
+        op_ids.push(tx.commit().op_id().clone());
+        std::thread::sleep(std::time::Duration::from_millis(1));
+    }
+    let repo = repo_loader.load_at_head(settings).unwrap();
+    // Test the setup. The assumption here is that the parent order matches the
+    // order in which they were merged (which currently matches the transaction
+    // commit order), so we want to know make sure they appear in a certain
+    // order, so the caller can decide the order by passing them to this
+    // function in a certain order.
+    assert_eq!(*repo.operation().parent_ids(), op_ids);
+    repo
+}
+
 pub fn read_file(store: &Store, path: &RepoPath, id: &FileId) -> Vec<u8> {
     let mut reader = store.read_file(path, id).unwrap();
     let mut content = vec![];
@@ -176,15 +213,20 @@ pub fn write_file(store: &Store, path: &RepoPath, contents: &str) -> FileId {
     store.write_file(path, &mut contents.as_bytes()).unwrap()
 }
 
-pub fn write_normal_file(tree_builder: &mut TreeBuilder, path: &RepoPath, contents: &str) {
+pub fn write_normal_file(
+    tree_builder: &mut TreeBuilder,
+    path: &RepoPath,
+    contents: &str,
+) -> FileId {
     let id = write_file(tree_builder.store(), path, contents);
     tree_builder.set(
         path.clone(),
         TreeValue::File {
-            id,
+            id: id.clone(),
             executable: false,
         },
     );
+    id
 }
 
 pub fn write_executable_file(tree_builder: &mut TreeBuilder, path: &RepoPath, contents: &str) {
@@ -203,7 +245,7 @@ pub fn write_symlink(tree_builder: &mut TreeBuilder, path: &RepoPath, target: &s
     tree_builder.set(path.clone(), TreeValue::Symlink(id));
 }
 
-pub fn create_tree(repo: &Arc<ReadonlyRepo>, path_contents: &[(&RepoPath, &str)]) -> Tree {
+pub fn create_single_tree(repo: &Arc<ReadonlyRepo>, path_contents: &[(&RepoPath, &str)]) -> Tree {
     let store = repo.store();
     let mut tree_builder = store.tree_builder(store.empty_tree_id().clone());
     for (path, contents) in path_contents {
@@ -213,15 +255,15 @@ pub fn create_tree(repo: &Arc<ReadonlyRepo>, path_contents: &[(&RepoPath, &str)]
     store.get_tree(&RepoPath::root(), &id).unwrap()
 }
 
+pub fn create_tree(repo: &Arc<ReadonlyRepo>, path_contents: &[(&RepoPath, &str)]) -> MergedTree {
+    MergedTree::legacy(create_single_tree(repo, path_contents))
+}
+
 #[must_use]
-pub fn create_random_tree(repo: &Arc<ReadonlyRepo>) -> TreeId {
-    let mut tree_builder = repo
-        .store()
-        .tree_builder(repo.store().empty_tree_id().clone());
+pub fn create_random_tree(repo: &Arc<ReadonlyRepo>) -> MergedTreeId {
     let number = rand::random::<u32>();
     let path = RepoPath::from_internal_string(format!("file{number}").as_str());
-    write_normal_file(&mut tree_builder, &path, "contents");
-    tree_builder.write_tree()
+    create_tree(repo, &[(&path, "contents")]).id()
 }
 
 pub fn create_random_commit<'repo>(
@@ -239,14 +281,23 @@ pub fn create_random_commit<'repo>(
         .set_description(format!("random commit {number}"))
 }
 
-pub fn dump_tree(store: &Arc<Store>, tree_id: &TreeId) -> String {
+pub fn dump_tree(store: &Arc<Store>, tree_id: &MergedTreeId) -> String {
     use std::fmt::Write;
     let mut buf = String::new();
-    writeln!(&mut buf, "tree {}", tree_id.hex()).unwrap();
-    let tree = store.get_tree(&RepoPath::root(), tree_id).unwrap();
+    writeln!(
+        &mut buf,
+        "tree {}",
+        tree_id
+            .to_merge()
+            .iter()
+            .map(|tree_id| tree_id.hex())
+            .join("&")
+    )
+    .unwrap();
+    let tree = store.get_root_tree(tree_id).unwrap();
     for (path, value) in tree.entries() {
-        match value {
-            TreeValue::File { id, executable: _ } => {
+        match value.into_resolved() {
+            Ok(Some(TreeValue::File { id, executable: _ })) => {
                 let file_buf = read_file(store, &path, &id);
                 let file_contents = String::from_utf8_lossy(&file_buf);
                 writeln!(
@@ -256,13 +307,10 @@ pub fn dump_tree(store: &Arc<Store>, tree_id: &TreeId) -> String {
                 )
                 .unwrap();
             }
-            TreeValue::Symlink(id) => {
+            Ok(Some(TreeValue::Symlink(id))) => {
                 writeln!(&mut buf, "  symlink {path:?} ({})", id.hex()).unwrap();
             }
-            TreeValue::Conflict(id) => {
-                writeln!(&mut buf, "  conflict {path:?} ({})", id.hex()).unwrap();
-            }
-            TreeValue::GitSubmodule(id) => {
+            Ok(Some(TreeValue::GitSubmodule(id))) => {
                 writeln!(&mut buf, "  submodule {path:?} ({})", id.hex()).unwrap();
             }
             entry => {
